from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import pandas as pd
from IPython.display import Markdown, display

from edvart.data_types import (
    DataType,
    infer_data_type,
    is_boolean,
    is_categorical,
    is_date,
    is_missing,
    is_numeric,
    is_unique,
)
from edvart.pandas_formatting import hide_index, render_dictionary, series_to_frame
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import ReportSection, Section, Verbosity


# pylint: disable=invalid-name
class OverviewSubsection(IntEnum):
    """
    Enum of possible subsections of the Overview section.
    """

    QuickInfo = 1
    DataTypes = 2
    DataPreview = 3
    MissingValues = 4
    RowsWithMissingValue = 5
    ConstantOccurrence = 6
    DuplicateRows = 7

    def __str__(self):
        return self.name


class Overview(ReportSection):
    """Generates the Overview section of the report.

    Contains an enum OverviewSubsection of possible subsections.

    Parameters
    ----------
    subsections : List[OverviewSubsection], optional
        List of subsections to include into the Overview section.
        All subsections in OverviewSubsection are used by default.
    verbosity : Verbosity
        Generated code verbosity global to the Overview sections
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do overview analysis. All columns are used by default.
    verbosity_quick_info : Verbosity, optional
        Quick info subsection code verbosity.
    verbosity_data_types : Verbosity, optional
        Data types subsection code verbosity.
    verbosity_data_preview : Verbosity, optional
        Data preview subsection code verbosity.
    verbosity_missing_values : Verbosity, optional
        Missing values subsection code verbosity.
    verbosity_rows_with_missing_value : Verbosity, optional
        Rows with missing value subsection code verbosity.
    verbosity_constant_occurrence : Verbosity, optional
        Constant values subsection code verbosity.
    verbosity_duplicate_rows : Verbosity, optional
        Duplicate rows subsection code verbosity.
    """

    # By default use all subsections
    _DEFAULT_SUBSECTIONS_TO_SHOW = list(OverviewSubsection)

    def __init__(
        self,
        subsections: Optional[List[OverviewSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        verbosity_quick_info: Optional[Verbosity] = None,
        verbosity_data_types: Optional[Verbosity] = None,
        verbosity_data_preview: Optional[Verbosity] = None,
        verbosity_missing_values: Optional[Verbosity] = None,
        verbosity_rows_with_missing_value: Optional[Verbosity] = None,
        verbosity_constant_occurrence: Optional[Verbosity] = None,
        verbosity_duplicate_rows: Optional[Verbosity] = None,
    ):
        # Propagate global verbosity to subsection verbosities
        verbosity_quick_info = verbosity_quick_info or verbosity
        verbosity_data_types = verbosity_data_types or verbosity
        verbosity_data_preview = verbosity_data_preview or verbosity
        verbosity_missing_values = verbosity_missing_values or verbosity
        if verbosity_rows_with_missing_value is None:
            verbosity_rows_with_missing_value = verbosity
        verbosity_constant_occurrence = verbosity_constant_occurrence or verbosity
        verbosity_duplicate_rows = verbosity_duplicate_rows or verbosity

        # Store subsection verbosities
        self.subsection_verbosities = {
            OverviewSubsection.QuickInfo: verbosity_quick_info,
            OverviewSubsection.DataTypes: verbosity_data_types,
            OverviewSubsection.DataPreview: verbosity_data_preview,
            OverviewSubsection.MissingValues: verbosity_missing_values,
            OverviewSubsection.RowsWithMissingValue: verbosity_rows_with_missing_value,
            OverviewSubsection.ConstantOccurrence: verbosity_constant_occurrence,
            OverviewSubsection.DuplicateRows: verbosity_duplicate_rows,
        }

        if subsections is None:
            self.subsections_to_show = self._DEFAULT_SUBSECTIONS_TO_SHOW
        else:
            self.subsections_to_show = subsections

        self.subsections_to_show_with_low_verbosity = [
            sub
            for sub in self.subsections_to_show
            if self.subsection_verbosities[sub] == Verbosity.LOW
        ]

        # Construct objects that implement subsections
        enum_to_implementation = {
            OverviewSubsection.QuickInfo: QuickInfo(verbosity_quick_info, columns),
            OverviewSubsection.DataTypes: DataTypes(verbosity_data_types, columns),
            OverviewSubsection.DataPreview: DataPreview(verbosity_data_preview, columns),
            OverviewSubsection.MissingValues: MissingValues(verbosity_missing_values, columns),
            OverviewSubsection.RowsWithMissingValue: RowsWithMissingValue(
                verbosity_rows_with_missing_value, columns
            ),
            OverviewSubsection.ConstantOccurrence: ConstantOccurrence(
                verbosity_constant_occurrence, columns
            ),
            OverviewSubsection.DuplicateRows: DuplicateRows(verbosity_duplicate_rows, columns),
        }

        subsections_implementations = [
            enum_to_implementation[sub] for sub in self.subsections_to_show
        ]
        super().__init__(subsections_implementations, verbosity, columns)

    @property
    def name(self) -> str:
        return "Overview"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity == Verbosity.LOW:
            imports = {"from edvart.report_sections.dataset_overview import show_overview"}
            if self.subsections_to_show_with_low_verbosity != self._DEFAULT_SUBSECTIONS_TO_SHOW:
                imports.add(
                    "from edvart.report_sections.dataset_overview import OverviewSubsection"
                )

            for subsec in self.subsections:
                if subsec.verbosity > Verbosity.LOW:
                    imports.update(subsec.required_imports())

            return list(imports)
        return super().required_imports()

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.LOW:
            code = "show_overview(df=df"
            if self.subsections_to_show_with_low_verbosity != self._DEFAULT_SUBSECTIONS_TO_SHOW:
                arg_subsections_names = [
                    f"OverviewSubsection.{str(sub)}"
                    for sub in self.subsections_to_show_with_low_verbosity
                ]
                code += f", subsections={arg_subsections_names}".replace("'", "")
            if self.columns is not None:
                code += f", columns={self.columns}"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))
            for subsec in self.subsections:
                if subsec.verbosity > Verbosity.LOW:
                    subsec.add_cells(cells, df=df)
        else:
            super().add_cells(cells, df=df)

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)


def show_overview(
    df: pd.DataFrame,
    subsections: Optional[List[OverviewSubsection]] = None,
    columns: Optional[List[str]] = None,
) -> None:
    """Generates overview analysis for df.

    Parameters
    ----------
    df : pd.DataFrame
        Data to be analyzed
    subsections : List[OverviewSubsection], optional
        Subsections to include into the overview
    columns : List[str], optional
        Subset of columns of df to consider in overview, by default all columns are used.
    """
    if columns is not None:
        df = df[columns]
    overview = Overview(subsections=subsections, verbosity=Verbosity.LOW, columns=columns)
    for sub in overview.subsections:
        sub.show(df)


class QuickInfo(Section):
    """Generates the Quick info subsection.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns to consider in quick info.
        If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Quick Info"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.dataset_overview import quick_info"]
        return []

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "quick_info(df=df)"
        else:
            default_call = f"quick_info(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = (
                get_code(render_dictionary)
                + 2 * "\n"
                + get_code(quick_info)
                + 2 * "\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Renders the quick info table in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data which to use for the table.
        """
        display(Markdown(self.get_title(section_level=2)))
        quick_info(df, columns=self.columns)


def quick_info(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    additional_rows: Optional[Dict[str, Any]] = None,
) -> None:
    """Renders a quick info table about df in the calling notebook.

    Parameters
    ----------
    df : pd.DataFrame
        Data which to analyze.
    columns : List[str], optional
        List of columns of df to analyze. All columns of df are used by default.
    additional_rows : Dict[str, Any], optional
        Additional custom rows to add to the table.
    """
    if columns is not None:
        df = df[columns]
    missing_cells = df.isna().sum().sum()
    missing_cells_percent = 100 * missing_cells / (df.shape[0] * df.shape[1])

    zeros = (df == 0).sum().sum()
    zeros_percent = 100 * zeros / (df.shape[0] * df.shape[1])

    duplicate_rows = df.duplicated().sum()
    duplicate_rows_percent = 100 * duplicate_rows / len(df)

    df_info_rows = {
        "Rows": df.shape[0],
        "Columns": df.shape[1],
        "Missing cells": f"{missing_cells} ({missing_cells_percent:,.02f} %)",
        "Zeros": f"{zeros} ({zeros_percent:.02f} %)",
        "Duplicate rows": f"{duplicate_rows} ({duplicate_rows_percent:,.02f} %)",
    }

    if additional_rows is not None:
        df_info_rows = {**df_info_rows, **additional_rows}

    render_dictionary(df_info_rows)


class DataTypes(Section):
    """Generates data types inference subsection.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns for which to infer data type.
        If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Data Types"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        base_imports = ["from edvart.pandas_formatting import hide_index"]
        if self.verbosity <= Verbosity.MEDIUM:
            return base_imports + ["from edvart.report_sections.dataset_overview import data_types"]
        return base_imports + [
            "from enum import IntEnum",
            "import numpy as np",
            "from IPython.display import display",
            "import warnings",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds data type inference cells to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "data_types(df=df)"
        else:
            default_call = f"data_types(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = "\n\n".join(
                (
                    *(
                        get_code(function_or_class)
                        for function_or_class in (
                            series_to_frame,
                            DataType,
                            is_unique,
                            is_numeric,
                            is_missing,
                            is_categorical,
                            is_boolean,
                            is_date,
                            infer_data_type,
                            data_types,
                        )
                    ),
                    default_call,
                )
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Renders a table with inferred data types in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame:
            Data for which to infer data types.
        """
        display(Markdown(self.get_title(section_level=2)))
        data_types(df=df, columns=self.columns)


def data_types(df: pd.DataFrame, columns: Optional[List[str]] = None) -> None:
    """Renders a table with inferred data types in the calling notebook.

    Parameters
    ----------
    df : pd.DataFrame:
        Data for which to infer data types
    columns : List[str], optional
        List of columns for which to infer data type. All columns of df are used by default.
    """
    if columns is not None:
        df = df[columns]
    dtypes = df.apply(
        func=lambda x_: str(infer_data_type(x_)),
        axis=0,
        result_type="expand",
    )

    # Convert result to frame for viewing
    dtypes_frame = series_to_frame(
        series=dtypes, index_name="Column Name", column_name="Inferred Data Type"
    )

    display(hide_index(dtypes_frame))


class DataPreview(Section):
    """Generates data preview (head, tail, sample) subsection.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns to preview.
        If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Data Preview"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.dataset_overview import data_preview"]
        return [
            "from IPython.display import display",
            "from IPython.display import Markdown",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds dataframe preview cells to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "data_preview(df=df)"
        else:
            default_call = f"data_preview(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = get_code(data_preview) + 2 * "\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Renders data preview tables in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data which to preview
        """
        display(Markdown(self.get_title(section_level=2)))
        data_preview(df=df, columns=self.columns)


def data_preview(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    n_head: int = 5,
    n_tail: int = 5,
    n_sample: int = 5,
) -> None:
    """Renders data preview tables in the calling notebook.

    Parameters
    ----------
    df : pd.DataFrame
        Data which to preview.
    columns : List[str], optional
        Columns of df to preview. All columns of df are used by default.
    n_head : int
        Number of first n rows of df to render, if None no preview is rendered.
    n_tail : int
        Number of last n rows of df to render, if None no preview is rendered.
    n_sample : int
        Size of random sample of df to render, if None no preview is rendered.
    """
    if columns is not None:
        df = df[columns]
    if n_head is not None:
        display(Markdown("### First rows"))
        display(df.head(n_head))
    if n_tail is not None:
        display(Markdown("### Last rows"))
        display(df.tail(n_tail))
    if n_sample is not None:
        display(Markdown("### Sample"))
        display(df.sample(min(n_sample, len(df))))


class MissingValues(Section):
    """Generates missing values percentages table for each column of the dataframe.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns for which to count missing values. If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Missing Values"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        base_imports = ["from edvart.pandas_formatting import hide_index"]
        if self.verbosity <= Verbosity.MEDIUM:
            return base_imports + [
                "from edvart.report_sections.dataset_overview import missing_values"
            ]
        return base_imports + [
            "from IPython.display import display",
            "import matplotlib.pyplot as plt",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds code cells which calculate missing values percentage table to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "missing_values(df=df)"
        else:
            default_call = f"missing_values(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = (
                get_code(series_to_frame)
                + 2 * "\n"
                + get_code(missing_values)
                + 2 * "\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates missing values percentages table in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        missing_values(df=df, columns=self.columns)


def missing_values(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    bar_plot: bool = True,
    bar_plot_figsize: Tuple[int, int] = (15, 6),
    bar_plot_title: str = "Missing Values Percentage of Each Column",
    bar_plot_ylim: float = 0,
    bar_plot_color: str = "#FFA07A",
    **bar_plot_args: Dict[str, Any],
) -> None:
    """Displays a table of missing values percentages for each column of df and a bar plot
    of the percentages.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe for which to calculate missing values.
    columns : Optional[List[str]], optional
        Subset of columns for which to calculate missing values percentage.
        If None, all columns of df are used.
    bar_plot : bool (default = False)
        Whether to also display a bar plot visualizing missing values percentages for each
        column.
    bar_plot_figsize : Tuple[int, int]
        Width and height of the bar plot.
    bar_plot_title : str
        Title of the bar plot.
    bar_plot_ylim : float
        Bar plot y axis bottom limit.
    bar_plot_color : str
        Color of bars in the bar plot in hex format.
    bar_plot_args : Dict[str, Any]
        Additional kwargs passed to pandas.Series.bar.
    """
    if columns is not None:
        df = df[columns]

    # Count null values
    null_count = df.isna().sum()
    null_percentage = 100 * null_count / len(df)

    if null_count.sum() == 0:
        print("There are no missing values")
        return

    # Convert series to frames
    null_count_frame = series_to_frame(
        series=null_count, index_name="Column Name", column_name="Null Count"
    )
    null_percentage_frame = series_to_frame(
        series=null_percentage, index_name="Column Name", column_name="Null %"
    )
    # Merge null count and percentage into one frame
    null_stats_frame = null_count_frame.merge(null_percentage_frame, on="Column Name").sort_values(
        "Null Count", ascending=False
    )

    display(
        hide_index(null_stats_frame)
        .bar(color="#FFA07A", subset=["Null %"], vmax=100)
        .format({"Null %": "{:.03f}"})
    )

    # Display bar plot of missing value percentages
    if bar_plot:
        (
            null_percentage_frame.sort_values("Null %", ascending=False)
            .plot.bar(
                x="Column Name",
                figsize=bar_plot_figsize,
                title=bar_plot_title,
                ylim=bar_plot_ylim,
                color=bar_plot_color,
                **bar_plot_args,
            )
            .set_ylabel("Missing Values [%]")
        )
        plt.show()


class ConstantOccurrence(Section):
    """Generates table with occurrence of a constant in each column.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns to count constant occurrence in. If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Constant Occurrence"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        base_imports = ["from edvart.pandas_formatting import hide_index"]
        if self.verbosity <= Verbosity.MEDIUM:
            return base_imports + [
                "from edvart.report_sections.dataset_overview import constant_occurrence"
            ]
        return base_imports + ["from IPython.display import display"]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds code cells which calculate constant occurrence table to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "constant_occurrence(df=df)"
        else:
            default_call = f"constant_occurrence(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = (
                get_code(series_to_frame)
                + 2 * "\n"
                + get_code(constant_occurrence)
                + 2 * "\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates constant occurrence table in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        constant_occurrence(df=df, columns=self.columns)


def constant_occurrence(
    df: pd.DataFrame, columns: Optional[List[str]] = None, constant: Any = 0
) -> None:
    """Displays a table with occurrence of a constant in each column.

    By default, check for 0 occurrence.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe for which to calculate constant values occurrence.
    columns : Optional[List[str]], optional
        Subset of columns for which to calculate constant values occurrence.
        If None, all columns of df are used.
    constant : Any
        Constant for which to check occurrence in df, by default 0.
    """
    if columns is not None:
        df = df[columns]

    # Count constant counts
    constant_count = (df == constant).sum()
    constant_percentage = 100 * constant_count / len(df)

    constant_formatted = f"<i>{constant!r}</i>"
    constant_count_name = f"{constant_formatted} Count"
    constant_perc_name = f"{constant_formatted} %"

    # Convert series to frames
    constant_count_frame = series_to_frame(
        series=constant_count,
        index_name="Column Name",
        column_name=constant_count_name,
    )

    constant_percentage_frame = series_to_frame(
        series=constant_percentage,
        index_name="Column Name",
        column_name=constant_perc_name,
    )

    # Merge absolute and relative counts
    constant_stats_frame = constant_count_frame.merge(
        constant_percentage_frame, on="Column Name"
    ).sort_values(constant_perc_name, ascending=False)

    # Display table
    display(
        hide_index(constant_stats_frame)
        .bar(color="#FFA07A", subset=[constant_perc_name], vmax=100)
        .format({constant_perc_name: "{:.03f}"})
    )


class RowsWithMissingValue(Section):
    """Counts number of rows with at least one value missing.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns to consider when counting. If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Rows With Missing Value"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.dataset_overview import missing_value_row_count"]
        return ["from IPython.display import display"]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds code cells which count the number of rows with missing value to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "missing_value_row_count(df=df)"
        else:
            default_call = f"missing_value_row_count(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = (
                get_code(render_dictionary)
                + 2 * "\n"
                + get_code(missing_value_row_count)
                + 2 * "\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates a table with missing value row count and percentage in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        missing_value_row_count(df=df, columns=self.columns)


def missing_value_row_count(df: pd.DataFrame, columns: Optional[List[str]] = None) -> None:
    """Displays a table with missing value row count and percentage.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe for which to count missing value rows.
    columns : Optional[List[str]], optional
        List of columns to consider when counting. If None, all columns are used.
    """
    if columns is not None:
        df = df[columns]

    # Count rows with at least one value missing
    num_rows_missing_value = df.isna().any(axis=1).sum()

    if num_rows_missing_value == 0:
        print("There are no missing values")
        return

    # Relative count
    percentage_rows_missing_value = 100 * num_rows_missing_value / len(df)

    missing_value_rows_info = {
        "Missing value column subset": "all columns" if columns is None else columns,
        "Missing value row count": num_rows_missing_value,
        "Missing value row percentage": f"{percentage_rows_missing_value:.02f} %",
    }

    # Display table
    render_dictionary(missing_value_rows_info)


class DuplicateRows(Section):
    """Counts number of duplicated rows.

    Parameters
    ----------
    verbosity : Verbosity
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns to consider when counting. If None, all columns are used.
    """

    @property
    def name(self) -> str:
        return "Duplicate Rows"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.dataset_overview import duplicate_row_count"]
        return ["from IPython.display import display"]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds code cells which count the number of duplicated rows to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        if self.columns is None:
            default_call = "duplicate_row_count(df=df)"
        else:
            default_call = f"duplicate_row_count(df=df, columns={self.columns})"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        elif self.verbosity == Verbosity.HIGH:
            code = (
                get_code(render_dictionary)
                + 2 * "\n"
                + get_code(duplicate_row_count)
                + 2 * "\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Displays a table with duplicated row count and percentage in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        duplicate_row_count(df=df, columns=self.columns)


def duplicate_row_count(df: pd.DataFrame, columns: Optional[List[str]] = None) -> None:
    """Displays a table with duplicated row count and percentage.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe for which to count missing value rows.
    columns : Optional[List[str]], optional
        List of columns to consider when counting. If None, all columns are used.
    """
    if columns is not None:
        df = df[columns]

    # Count duplicated rows
    num_duplicated_rows = df.duplicated().sum()

    if num_duplicated_rows == 0:
        print("There are no duplicated rows")
        return

    # Relative count
    percentage_duplicated_rows = 100 * num_duplicated_rows / len(df)

    duplicate_rows_info = {
        "Duplicate rows column subset": "all columns" if columns is None else columns,
        "Duplicate row count": num_duplicated_rows,
        "Duplicate row percentage": f"{percentage_duplicated_rows:.02f} %",
    }

    # Display table
    render_dictionary(duplicate_rows_info)



from inspect import getsource
from itertools import dropwhile
from textwrap import dedent
from typing import Any


def total_dedent(input_string: str) -> str:
    """Removes all white space from the beginning of each line.

    Parameters
    ----------
    input_string : str
        Input string with lines.

    Returns
    -------
    str
        input_string with no whitespace at the beginning of each line.
    """
    input_string = input_string.strip()
    lstripped_lines = [l.strip() for l in input_string.split("\n")]
    return "\n".join(lstripped_lines)


def code_dedent(input_string: str) -> str:
    """Removes all white spaces from each line that is common for all lines.

    Parameters
    ----------
    input_string : str
        Input string with lines.

    Returns
    -------
    str
        input_string with common leading whitespace removed from each line.
    """

    return dedent(input_string.strip("\n"))


def dedecorate(input_string: str) -> str:
    """Removes all decorators from the beginning of a function source.

    Parameters
    ----------
    input_string : str
        Input function source.

    Returns
    -------
    str
        input_string with beginning lines starting with '@' removed.
    """
    lines = input_string.splitlines()
    filtered_lines = dropwhile(lambda l: l.lstrip().startswith("@"), lines)

    return "\n".join(filtered_lines)


def get_code(code_object: Any) -> str:
    """Gets the source code of code object and formats it.

    Parameters
    ----------
    code_object : Any
        Object from which to extract code (function, method)

    Returns
    -------
    str
        Formatted code
    """
    code = getsource(code_object)
    code = dedecorate(code)
    # Dedent code
    code = code_dedent(code)

    return code



import os
from contextlib import contextmanager
from typing import Any, Dict, Iterable, Iterator, List, Literal, Optional, Tuple, Union

import pandas as pd
import plotly
import statsmodels.api as sm
from scipy import stats

from edvart.data_types import is_numeric


def top_frequent_values(series: pd.Series, n_top: int = 10) -> Dict[Any, float]:
    """
    Counts top n most frequent values in series along with other value counts and NULL value counts.

    Parameters
    ---
    series: pd.Series
        Input series of data for which frequencies will be calculated
    n_top: int
        Number of values for which actual frequencies will be counted, other values will be grouped
        into 'Other' category

    Returns
    -------
    result_dict: Dict
        Dictionary with the mapping {'value': 'frequency (relative frequency)'}
    """
    # Calculate frequencies
    counts = series.value_counts()
    nan_count = series.isna().sum()
    result_dict = {
        **(counts[:n_top].to_dict()),
        "Other values count": counts[n_top:].sum(),
        "Null": nan_count,
    }

    # Add relative frequencies
    for key, _value in result_dict.items():
        result_dict[key] = f"{result_dict[key]:,} ({100 * result_dict[key] / len(series):.02f} %)"

    return result_dict


def reindex_to_datetime(
    df: pd.DataFrame,
    datetime_column: str,
    keep_index: Optional[str] = None,
    unit: str = "ns",
    origin: Union[str, pd.Timestamp] = "unix",
    sort: bool = True,
) -> pd.DataFrame:
    """Reindex a given DataFrame to be indexed by a pd.DateTimeIndex.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to reindex.
    datetime_column : str
        Which column containing datetimes to index by.
    keep_index : str, optional
        Name of column to store the original index. The original index will be discarded
        by default.
    unit : str (default = "ns")
        Numeric values would be parsed as number of units from origin.
    origin : Union[str, pd.Timestamp] (default = "unix")
        Define the reference date. Numeric values would be parsed as number of units
        (defined by unit) since this reference date. By default unix epoch 1970-01-01.
    sort : bool (default = True)
        Whether to sort according to the index.

    Returns
    -------
    pd.DataFrame
        Reindexed df.
    """
    df = df.copy()
    new_index = pd.to_datetime(df[datetime_column], unit=unit, origin=origin)
    if keep_index is not None:
        df[keep_index] = df.index
    df = df.drop(datetime_column, axis="columns")
    df.index = pd.DatetimeIndex(new_index)
    if sort:
        df = df.sort_index()

    return df


def reindex_to_period(
    df: pd.DataFrame,
    period_column: str,
    freq: Optional[str],
    keep_index: Optional[str] = None,
    sort: bool = True,
) -> pd.DataFrame:
    """Reindex a given DataFrame to be indexed by a pd.PeriodIndex.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to reindex.
    period_column : str
        Which column containing periods to index by.
    freq : Union[str, pd.Offset]
        One of pandas' offset strings or an Offset object. Will be inferred by default.
    keep_index : str, optional
        Name of column to store the original index. The original index will be discarded
        by default.
    sort : bool (default = True)
        Whether to sort according to the index.

    Returns
    -------
    pd.DataFrame
        Reindexed df.
    """
    df = df.copy()
    new_index = df[period_column].to_period(freq=freq)
    if keep_index is not None:
        df[keep_index] = df.index
    df = df.drop(period_column, axis="columns")
    df.index = pd.PeriodIndex(new_index)
    if sort:
        df = df.sort_index()

    return df


def hsl_wheel_colorscale(n: int, saturation=0.5, lightness=0.5) -> Iterable[str]:
    """
    Generate a colorscale of n discrete colors.

    Colors are equally spaced around the complete HSL wheel with constant saturation and lightness.

    Returns
    -------
    Iterable[str]
        An iterable of n plotly-compatible HSL strings.
    """
    for i in range(n):
        yield f"hsl({(i / n) * 360 :.2f}, {saturation * 100 :.2f}%, {lightness * 100 :.2f}%)"


def make_discrete_colorscale(colorscale: List[str], n_colors: int) -> Iterable[Tuple[float, str]]:
    """
    Generate a colorscale of n discrete colors for use in `plotly.graph_objects`.

    Note that when using `plotly.express`, the parameter `color_discrete_sequence`
    can be used instead.

    Parameters
    ----------
    colorscale : List[str]
        A list of plotly-compatible colors.
    n_colors : int
        Number of colors to in the generated colorscale.

    Returns
    -------
    Iterable[Tuple[float, str]]
        An iterable of 2n tuples, where each tuple contains a value between 0 and 1
        (the values are equally spaced in the interval and each value appears twice), and one of the
        colors from the `colorscale`.

    Examples
    --------
    >>> list(make_discrete_colorscale(["red", "green", "blue"], 4))
    [
        (0, "red"), (0.25, "red"),
        (0.25, "green"), (0.5, "green"),
        (0.5, "blue"), (0.75, "blue"),
        (0.75, "red"), (1, "red")
    ]
    """
    for i in range(n_colors):
        color = colorscale[i % len(colorscale)]
        yield (i / n_colors, color)
        yield ((i + 1) / n_colors, color)


def get_default_discrete_colorscale(n_colors: int) -> List[Tuple[float, str]]:
    """
    Get a default Plotly-compatible colorscale of n discrete colors.

    Parameters
    ----------
    n_colors : int
        Number of colors.

    Returns
    -------
    list[tuple[float, str]]
        A list of 2n tuples, where each tuple contains a value between 0 and 1 and a
        plotly-compatible color string.
    """
    if n_colors <= 10:
        colorscale = plotly.colors.qualitative.Plotly
    elif n_colors <= 24:
        colorscale = plotly.colors.qualitative.Light24
    else:
        colorscale = list(hsl_wheel_colorscale(n_colors))
    return list(make_discrete_colorscale(colorscale, n_colors))


def select_numeric_columns(df: pd.DataFrame, columns: Optional[List[str]]) -> List[str]:
    """
    Select all numeric columns from a DataFrame if `columns` is `None`,
    or check if all specified columns are numeric if `columns` is a list of column names.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to select or check columns from.
    columns : List[str], optional
        Specified columns.

    Returns
    -------
    List[str]
        List of numeric or specified columns

    Raises
    ------
    ValueError
        If a non-numeric column is specified in `columns`.
    """
    # By default use only numeric columns
    if columns is None:
        return [col for col in df.columns if is_numeric(df[col])]
    for col in columns:
        if not is_numeric(df[col]):
            raise ValueError(f"Cannot use non-numeric column {col} of dtype {df[col].dtype}.")
    return columns


#######################
# Statistic functions #
#######################


def num_unique_values(series: pd.Series) -> int:
    """
    Return number of unique values.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    int
    """
    return series.nunique()


def median_absolute_deviation(series: pd.Series) -> float:
    """
    Return median absolute deviation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return median((series - series.mean()).abs())


def coefficient_of_variation(series: pd.Series) -> float:
    """
    Return coefficient of variation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.std() / series.mean()


def minimum(series: pd.Series) -> float:
    """
    Return minimum.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.min()


def maximum(series: pd.Series) -> float:
    """
    Return maximum.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.max()


def quartile1(series: pd.Series) -> float:
    """
    Return first quartile.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.quantile(0.25)


def quartile3(series: pd.Series) -> float:
    """
    Return third quartile.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.quantile(0.75)


def mean(series: pd.Series) -> float:
    """
    Return mean.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.mean()


def median(series: pd.Series) -> float:
    """
    Return median.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.median()


def iqr(series: pd.Series) -> float:
    """
    Return inter quartile range.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.quantile(0.75) - series.quantile(0.25)


def value_range(series: pd.Series) -> float:
    """
    Return value range.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.max() - series.min()


def mode(series: pd.Series) -> float:
    """
    Return mode.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
        The most frequent value. `float('nan')` if the series contains only null values.
    """
    most_frequent = series.mode(dropna=True)
    if len(most_frequent) == 0:
        return float("nan")
    return most_frequent[0]


def std(series: pd.Series) -> float:
    """
    Return standard deviation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return series.std()


def mad(series: pd.Series) -> Any:
    """
    Return mean absolute deviation.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return (series - series.mean()).abs().mean()


def kurtosis(series: pd.Series) -> Any:
    """
    Return kurtosis.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return stats.kurtosis(series)


def skewness(series: pd.Series) -> Any:
    """
    Return skewness.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    if series.isnull().all():
        return float("nan")
    return stats.skew(series)


def sum_(series: pd.Series) -> float:
    """
    Return sum.

    Parameters
    ----------
    series: pd.Series
        Series on which the stat should be calculated.

    Returns
    -------
    float
    """
    return series.sum(min_count=1)


def pearson(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return pearson correlation coefficient.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame on which the stat should be calculated.

    Returns
    -------
    pd.DataFrame
    """
    return _corr(df, "pearson")


def spearman(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return spearman correlation coefficient.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame on which the stat should be calculated.

    Returns
    -------
    pd.DataFrame
    """
    return _corr(df, "spearman")


def kendall(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return kendall correlation coefficient.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame on which the stat should be calculated.

    Returns
    -------
    pd.DataFrame
    """
    return _corr(df, "kendall")


def _corr(df: pd.DataFrame, method: Literal["pearson", "kendall", "spearman"]) -> pd.DataFrame:
    return df.corr(method=method, numeric_only=True)


def contingency_table(df: pd.DataFrame) -> pd.DataFrame:
    """
    Return contingency table.

    Parameters
    ----------
    df: pd.DataFrame

    Returns
    -------
    pd.DataFrame
    """
    table = sm.stats.Table.from_data(df)
    return table.table_orig.astype(int)


@contextmanager
def env_var(name: str, value: str) -> Iterator[None]:
    """
    Set an environment variable for the duration of the context.

    Parameters
    ----------
    name : str
        Name of the environment variable.
    value : str
        Value of the environment variable.
    """
    original_env = os.environ.copy()
    os.environ[name] = value
    try:
        yield
    finally:
        os.environ = original_env



import warnings
from functools import partial
from typing import Any, Callable, Dict, List, Optional

import nbformat.v4 as nbfv4
import pandas as pd
import statsmodels.api as sm
from IPython.display import Markdown, display

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.pandas_formatting import format_number
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class StationarityTests(Section):
    """Generates the stationarity tests subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    """

    @property
    def name(self) -> str:
        return "Stationarity tests"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                "from edvart.report_sections.timeseries_analysis.stationarity_tests"
                " import show_stationarity_tests"
            ]
        return [
            "from IPython.display import display, Markdown",
            "import statsmodels.api as sm",
            "from functools import partial",
            "from edvart.data_types import is_numeric",
            "from edvart.pandas_formatting import format_number",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)
        default_call = "show_stationarity_tests(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(default_stationarity_tests)
                + "\n\n"
                + get_code(show_stationarity_tests)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates seasonal decomposition plot(s) in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_stationarity_tests(df=df, columns=self.columns)


def default_stationarity_tests() -> Dict[pd.Series, Callable[[pd.Series], "test_result"]]:
    """Return a dictionary of stationarity test and functions.

    Stationarity tests are:

        KPSS (constant)
            which has a null hypothesis that a given series is stationary around a
            constant value

        KPSS (trend)
            which has a null hypothesis that a given series is stationary around a
            constant-slope, i.e. a linear function

        Augmented Dickey-Fuller
            which has a null hypothesis of a unit root, i.e. non-stationarity.

    Returns
    -------
    Dict[str, Callable]
        A dictionary from test name to function.
    """
    return {
        "KPSS Test (constant)": partial(sm.tsa.stattools.kpss, regression="c", nlags="auto"),
        "KPSS Test (trend)": partial(sm.tsa.stattools.kpss, regression="ct", nlags="auto"),
        "Augmented Dickey-Fuller": sm.tsa.stattools.adfuller,
    }


@check_index_time_ascending
def show_stationarity_tests(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    kpss_const: bool = True,
    kpss_trend: bool = True,
    adfuller: bool = True,
) -> None:
    """Show stationarity for each numeric column.

    Parameters
    ----------
    df : pd.DataFrame
        Data to test.
    columns : List[str], optional
        List of columns to test. Only numeric columns can be used.
        All numeric columns are used by default.
    kpss_const : bool (default = True)
        Whether to perform KPSS (constant) test.
    kpss_trend : bool (default = True)
        Whether to perform KPSS (trend) test.
    adfuller : bool (default = True)
        Whether to perform Augmented Dickey-Fuller test.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    df = df.dropna()
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    stat_tests = default_stationarity_tests()
    if not kpss_const:
        stat_tests.pop("KPSS Test (constant)", None)
    if not kpss_trend:
        stat_tests.pop("KPSS Test (trend)", None)
    if not adfuller:
        stat_tests.pop("Augmented Dickey-Fuller Test", None)

    for col in columns:
        if not is_numeric(df[col]):
            raise ValueError(f"Cannot test stationarity of non-numeric column `{col}`")
        test_values_df = pd.DataFrame()
        display(Markdown(f"---\n### {col}"))
        for name, func in stat_tests.items():
            with warnings.catch_warnings(record=True) as warns:
                test_vals = func(df[col])
            stat, pvalue = test_vals[:2]
            value_dict = {
                "Test statistic": format_number(stat, thousand_separator=" "),
                "P-value": ("<" if len(warns) > 0 else "")
                + format_number(pvalue, thousand_separator=" "),
            }
            value_series = pd.Series(value_dict)
            test_values_df[name] = value_series
        display(test_values_df.style)



from .autocorrelation import Autocorrelation
from .boxplots_over_time import BoxplotsOverTime
from .fourier_transform import FourierTransform
from .rolling_statistics import RollingStatistics
from .seasonal_decomposition import SeasonalDecomposition
from .short_time_ft import ShortTimeFT
from .stationarity_tests import StationarityTests
from .time_series_line_plot import TimeSeriesLinePlot
from .timeseries_analysis import (  # pylint: disable=cyclic-import
    TimeseriesAnalysis,
    TimeseriesAnalysisSubsection,
)



from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
from IPython.display import Markdown, display

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class FourierTransform(Section):
    """Generates the Discrete Fourier Transform spectrum plot subsection.

    Parameters
    ----------
    sampling_rate : int
        The time series will be considered as samples from a lower-frequency at this rate, i.e.
        frequencies in multiples of (1 / sampling rate) will be analyzed.
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    """

    def __init__(
        self,
        sampling_rate: int,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
    ):
        if sampling_rate <= 0:
            raise ValueError(f"sampling_rate has to be a positive integer, not {sampling_rate}")
        super().__init__(verbosity, columns)
        self.sampling_rate = sampling_rate

    @property
    def name(self) -> str:
        return "Fourier Transform"

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)
        default_call = f"show_fourier_transform(df=df, sampling_rate={self.sampling_rate}"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = get_code(show_fourier_transform) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates Fourier transform spectrum plot(s) in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_fourier_transform(df=df, sampling_rate=self.sampling_rate, columns=self.columns)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                """from edvart.report_sections.timeseries_analysis.fourier_transform import (
                    FourierTransform, fourier_transform
                )"""
            ]
        return [
            "import numpy as np",
            "import matplotlib.pyplot as plt",
            "from edvart.data_types import is_numeric",
        ]


@check_index_time_ascending
def show_fourier_transform(
    df: pd.DataFrame,
    sampling_rate: int,
    columns: Optional[List[str]] = None,
    figsize: Tuple[float, float] = (15, 6),
    log: bool = False,
    freq_min: Optional[float] = None,
    freq_max: Optional[float] = None,
) -> None:
    """Generate Discrete Fourier Transform frequency vs amplitude plot.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    sampling_rate : int
        The time series will be considered as samples from a lower-frequency at this rate, i.e.
        frequencies in multiples of (1 / sampling rate) will be analyzed.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    figsize : Tuple[float, float] (default = (15, 6))
        Size of frequency-amplitude plot.
    log : bool (default = False)
        Whether to plot amplitude in logarithmic scale -- in decibel.
    freq_min : float, optional
        Lowest frequency to show in the plot. All computed frequencies are shown by default.
    freq_max : float, optional
        Highest frequency to show in the plot. All computed frequencies are shown by default.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    if sampling_rate <= 0:
        raise ValueError(f"sampling_rate has to be a positive integer, not {sampling_rate}")
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    else:
        for col in columns:
            if not is_numeric(df[col]):
                raise ValueError(f"Cannot perform Fourier transform for non-numeric column `{col}`")
    index_freq = pd.infer_freq(df.index) or ""
    for col in columns:
        # FFT requires samples at regular intervals
        df_col = df[col].interpolate(method="time")
        df_col_centered = df_col - df_col.mean()
        fft_result = np.fft.fft(df_col_centered)

        amplitude = np.abs(fft_result) * 2 / len(df)
        fft_freq = np.fft.fftfreq(len(amplitude), 1.0 / sampling_rate)
        idx_pos_freq = fft_freq > 0
        fft_freq, amplitude = fft_freq[idx_pos_freq], amplitude[idx_pos_freq]

        y = 10 * np.log10(amplitude) if log else amplitude
        _fig, ax = plt.subplots(figsize=figsize)
        ax.stem(fft_freq, y, markerfmt="")
        ax.set_xlabel(f"Frequency [1 / {sampling_rate}{index_freq}]")
        ax.set_ylabel("Amplitude" + (" [dB]" if log else ""))
        ax.set_xlim(freq_min, freq_max)
        display(Markdown(f"---\n### {col}"))
        plt.show()



from typing import Any, Callable, Dict, List, Optional, Union

import colorlover as cl
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from IPython.display import Markdown, display
from matplotlib import cm
from matplotlib.colors import Normalize, to_hex
from plotly.subplots import make_subplots

from edvart import utils
from edvart.data_types import DataType, infer_data_type
from edvart.report_sections.code_string_formatting import code_dedent, get_code
from edvart.report_sections.section_base import Section, Verbosity


class GroupAnalysis(Section):
    """Generate the group analysis section of the report.

    Parameters
    ----------
    groupby : Union[str, List[str]]
        Name of column or list of columns names to group by.
    verbosity : Verbosity (default = Verbosity.LOW)
        Generated code verbosity global to the Group analysis sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do group analysis. All columns are used by default.
    show_within_group_statistics : bool (default = True)
        Whether to show per-group statistics.
    show_group_missing_values : bool (default = True)
        Whether to show per-group missing values.
    show_group_distribution_plots : bool (default = True)
        Whether to show per-group distribution plots.

    Raises
    ------
    ValueError
        If groupby columns are not a subset of the columns of the input DataFrame df.
    """

    def __init__(
        self,
        groupby: Union[str, List[str]],
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        show_within_group_statistics: bool = True,
        show_group_missing_values: bool = True,
        show_group_distribution_plots: bool = True,
    ):
        if isinstance(groupby, str):
            groupby = [groupby]
        self.columns = columns
        self.groupby = groupby
        self.show_statistics = show_within_group_statistics
        self.show_missing_vals = show_group_missing_values
        self.show_dist = show_group_distribution_plots

        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Group Analysis"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"]
        """
        if self.verbosity == Verbosity.LOW:
            return ["from edvart.report_sections.group_analysis import show_group_analysis"]
        if self.verbosity == Verbosity.MEDIUM:
            return [
                code_dedent(
                    """from edvart.report_sections.group_analysis import (
                        default_group_descriptive_stats,
                        default_group_quantile_stats,
                        show_group_analysis,
                        group_barplot,
                        group_missing_values,
                        overlaid_histograms,
                        within_group_descriptive_stats,
                        within_group_quantile_stats,
                        within_group_stats
                    )"""
                )
            ]
        # verbosity HIGH
        return [
            "import colorlover as cl",
            "import matplotlib.pyplot as plt",
            "import matplotlib.cm as cm",
            "from matplotlib.colors import Normalize, to_hex",
            "import numpy as np",
            "import pandas as pd",
            "import plotly.graph_objects as go",
            "from edvart.data_types import infer_data_type, DataType",
            "from edvart import utils",
            "from typing import List, Dict, Optional, Callable",
            "from plotly.subplots import make_subplots",
        ]

    def _add_function_defs(self, cells: List[Dict[str, Any]]):
        """Add function definition cell to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        code = (
            get_code(default_group_quantile_stats)
            + "\n\n"
            + get_code(default_group_descriptive_stats)
            + "\n\n"
            + get_code(within_group_descriptive_stats)
            + "\n\n"
            + get_code(within_group_quantile_stats)
            + "\n\n"
            + get_code(within_group_stats)
            + "\n\n"
            + get_code(group_barplot)
            + "\n\n"
            + get_code(overlaid_histograms)
            + "\n\n"
            + get_code(group_missing_values)
        )

        cells.append(nbfv4.new_code_cell(code))

    def _add_cells_numeric_col(self, cells: List[Dict[str, Any]], column_name: str):
        """Add code cells for a numeric column at verbosity MEDIUM or HIGH.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        column_name : str
            Name of column for which to generate code.
        """
        code = ""
        if self.show_statistics:
            if self.verbosity == Verbosity.MEDIUM:
                code += (
                    f"within_group_stats(df=df, groupby={self.groupby}, column='{column_name}')\n"
                )
            else:
                code += code_dedent(
                    f"""
                            within_group_stats(
                                df=df,
                                groupby={self.groupby},
                                column='{column_name}',
                                stats=default_group_descriptive_stats()
                            )

                            within_group_stats(
                                df=df,
                                groupby={self.groupby},
                                column='{column_name}',
                                stats=default_group_quantile_stats()
                            )
                        """
                )

        if self.show_dist:
            code += f"overlaid_histograms(df=df, groupby={self.groupby}, column='{column_name}')"
        cells.append(nbfv4.new_code_cell(code))

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Add cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        if not set(self.groupby) <= set(df.columns):
            raise ValueError("Grouping by a column which is not in columns of df.")
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.LOW:
            if self.columns is None:
                code = f"show_group_analysis(df=df, groupby={self.groupby})"
            else:
                code = f"show_group_analysis(df=df, groupby={self.groupby}, columns={self.columns})"
            code_cell = nbfv4.new_code_cell(code)
            cells.append(code_cell)
            return

        if self.verbosity == Verbosity.HIGH:
            self._add_function_defs(cells)

        if self.show_missing_vals:
            cells.append(nbfv4.new_markdown_cell("## Missing values for each group"))
            if self.columns is None:
                code = f"group_missing_values(df=df, groupby={self.groupby})"
            else:
                code = (
                    f"group_missing_values(df=df, columns={self.columns}, groupby={self.groupby})"
                )
            cells.append(nbfv4.new_code_cell(code))

        columns = self.columns if self.columns is not None else df.columns

        if not self.show_statistics and not self.show_dist:
            return
        for col in columns:
            if col in self.groupby:
                continue
            cells.append(nbfv4.new_markdown_cell(f"## *{col}*"))
            datatype = infer_data_type(df[col])
            if datatype == DataType.NUMERIC:
                self._add_cells_numeric_col(cells, col)
            else:
                code = f"group_barplot(df=df, groupby={self.groupby}, column='{col}')"
                cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        columns = (
            self.columns
            if self.columns is not None
            else [col for col in df.columns if col not in self.groupby]
        )
        display(Markdown(self.get_title(section_level=1)))

        if self.show_missing_vals:
            display(Markdown("## Missing values for each group"))
            group_missing_values(df, self.groupby)
        if not self.show_dist and not self.show_statistics:
            return
        for col in columns:
            if col in self.groupby:
                continue
            display(Markdown(f"## *{col}*"))
            datatype = infer_data_type(df[col])
            if datatype == DataType.NUMERIC:
                if self.show_statistics:
                    within_group_stats(df, self.groupby, col)
                if self.show_dist:
                    overlaid_histograms(df, self.groupby, col)
            else:
                group_barplot(df, self.groupby, col)


def default_group_descriptive_stats() -> Dict[str, Callable[[pd.Series], float]]:
    """Descriptive statistic functions.

    Returns
    -------
    Dict[str, Callable[[pd.Series], float]]
        A dictionary of statistic function names and functions.
    """
    return {
        "# Unique values": utils.num_unique_values,
        "Sum": utils.sum_,
        "Mode": utils.mode,
        "Mean": utils.mean,
        "Std": utils.std,
        "Mean abs dev": utils.mad,
        "Median abs dev": utils.median_absolute_deviation,
        "Relative std": utils.coefficient_of_variation,
        "Kurtosis": utils.kurtosis,
        "Skewness": utils.skewness,
    }


def default_group_quantile_stats() -> Dict[str, Callable[[pd.Series], float]]:
    """Quantile statistic functions.

    Returns
    -------
    Dict[str, Callable[[pd.Series], float]]
        A dictionary of statistic function names and functions.
    """
    return {
        "Min": utils.minimum,
        "Q1": utils.quartile1,
        "Median": utils.median,
        "Q3": utils.quartile3,
        "Max": utils.maximum,
        "Range": utils.value_range,
        "IQR": utils.iqr,
    }


def within_group_descriptive_stats(
    df: pd.DataFrame, groupby: List[str], column: str, round_decimals: int = 2
):
    """Display within-group descriptive statistics for a column.

    Parameters
    ----------
    df : pd.DataFrame
        Data to display statistics for.
    groupby : List[str]
        List of column names to group data by.
    column : str
        Which column to display statistics for.
    round_decimals : int (default = 2)
        Number of decimals to round displayed results to.
    """
    within_group_stats(
        df=df,
        groupby=groupby,
        column=column,
        stats=default_group_descriptive_stats(),
        round_decimals=round_decimals,
    )


def within_group_quantile_stats(
    df: pd.DataFrame, groupby: List[str], column: str, round_decimals: int = 2
) -> None:
    """Display within-group quantile statistics for a column.

    Parameters
    ----------
    df : pd.DataFrame
        Data to display statistics for.
    groupby : List[str]
        List of column names to group data by.
    column : str
        Which column to display statistics for.
    round_decimals : int (default = 2)
        Number of decimals to round displayed results to.
    """
    within_group_stats(
        df=df,
        groupby=groupby,
        column=column,
        stats=default_group_quantile_stats(),
        round_decimals=round_decimals,
    )


def within_group_stats(
    df: pd.DataFrame,
    groupby: List[str],
    column: str,
    stats: Dict[str, Callable[[pd.Series], float]] = None,
    round_decimals: int = 2,
) -> None:
    """Display withing group statistics for a column of df grouped by one or other more columns.

    Parameters
    ----------
    df : pd.DataFrame
        Data to display statistics for.
    groupby : List[str]
        List of column names to group by.
    column : str
        Name of column to display statistics for.
    stats : Dict[str, Callable[[pd.Series], float]], optional
        A dictionary of statistic function names and functions.
        If None, default_group_quantile_stats() and default_group_descriptive_stats() will be
        used.
    round_decimals : int (default = 2)
        Number of decimals to round displayed results to.
    """
    if stats is None:
        within_group_descriptive_stats(df, groupby, column, round_decimals)
        within_group_quantile_stats(df, groupby, column, round_decimals)
        return

    df_grouped = df.groupby(groupby)[column]
    group_stats = []
    for name, func in stats.items():
        group_stats.append(df_grouped.apply(func).rename(name))
    stats_table = pd.concat(group_stats, axis=1)
    stats_table = stats_table.round(decimals=round_decimals)
    display(stats_table)


# pylint: disable=too-many-locals
def group_missing_values(
    df: pd.DataFrame,
    groupby: Union[str, List[str]],
    columns: Optional[List[str]] = None,
    round_decimals: int = 2,
    heatmap: bool = True,
    foreground_colormap: str = "bone",
    background_colormap: str = "OrRd",
    sort: bool = True,
    sort_by: Optional[List[str]] = None,
    ascending: bool = False,
) -> None:
    """Display per-group number and percentage of missing values in each column.

    Parameters
    ----------
    df : pd.DataFrame
        Data to display missing values for.
    groupby : str or List[str]
        Name of column or list of columns names to group by.
    columns : List[str], optional
        Subset of columns to analyze. All columns except those for grouping are used by default.
    round_decimals : int (default = 2)
        Number of decimals to round displayed results to.
    heatmap : bool (default = True)
        Whether to color missing value percentage cells according to the corresponding value.
    foreground_colormap : str (default = "bone")
        Color map of the foreground.
    background_colormap : str (default = "OrRd)
        Color map of the background.
    sort : bool (default = True)
        Whether to sort the results.
    sort_by : List[str], optional
        List of column names to sort the results by. Sort by all column by default.
    ascending : bool (default = False)
        If True, sort in ascending order, otherwise sort in descending order.

    Raises
    ------
    ValueError
        If groupby columns are not a subset of the columns of the input DataFrame df.
    """
    if isinstance(groupby, str):
        groupby = [groupby]
    if not set(groupby) <= set(df.columns):
        raise ValueError("Grouping by a column which is not in columns of df.")
    if columns is None:
        columns = [col for col in df.columns if col not in groupby]
    df_grouped = df.groupby(groupby)[columns]

    # Calculate number of samples in each group
    sizes = df_grouped.size().rename("Group Size")

    # Calculate missing values percentage of each column for each group
    missing = df_grouped.apply(lambda g: g.isna().sum(axis=0))
    missing = missing.divide(sizes, axis=0) * 100  # `/` divides by axis=-1
    missing = missing.fillna(value=0)
    missing = missing.round(decimals=round_decimals)

    if missing.sum().sum() == 0:
        print("There are no missing values")
        return

    # Concatenate group sizes and missing value percentages
    final_table = pd.concat([sizes, missing], axis=1)

    # Sort columns to better identify groups with missing data
    all_columns = [col for col in missing.columns if col not in groupby]
    if sort:
        if sort_by is None:
            sort_by = all_columns
        final_table = final_table.sort_values(by=sort_by, axis=0, ascending=ascending)

    # Drop columns with no missing data
    missing = final_table.sum(axis=0) != 0
    final_table = final_table.loc[:, missing]

    colored_columns = [col for col in final_table if col in all_columns]

    # Apply conditional formatting to each cell except group size column
    if heatmap:
        fg_cmap = cm.get_cmap(foreground_colormap)
        bg_cmap = cm.get_cmap(background_colormap)
        norm = Normalize(vmin=0, vmax=100)

        def color_cell(value):
            fg_hex = to_hex(fg_cmap(norm(value)))
            bg_hex = to_hex(bg_cmap(norm(value)))
            return f"""
                color: {fg_hex};
                background-color: {bg_hex};
            """

        render = final_table.style.applymap(
            func=color_cell, subset=pd.IndexSlice[:, colored_columns]
        ).format(formatter="{0:.2f} %", subset=pd.IndexSlice[:, colored_columns])
    else:
        render = final_table.style.format(
            formatter="{0:.2f} %", subset=pd.IndexSlice[:, colored_columns]
        )

    # Render table
    display(render)


def group_barplot(
    df: pd.DataFrame,
    groupby: List[str],
    column: str,
    group_count_threshold: int = 20,
    conditional_probability: bool = True,
    xaxis_tickangle: float = 0,
    alpha: float = 0.5,
):
    """Display a per-group barplot for a column.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    groupby : List[str]
        List of column names to group by.
    column : str
        Which column to analyze.
    group_count_threshold : int (default = 20)
        Maximum number of unique values in column to plot. If the number of unique values
        is higher, a warning will be issued and plot will not be shown.
    conditional_probability : bool (default = True)
        If True, conditional probability conditioned on group will be displayed,
        otherwise conditional frequency will be displayed.
    xaxis_tickangle : float (default = 0)
        Rotation angle of ticks on the x axis.
    alpha : float (default = 0.5)
        Opacity of bars in the plot.
    """
    num_cat = df[column].nunique()
    if num_cat > group_count_threshold:
        display(
            Markdown(
                f"Number of unique values ({num_cat}) greater than threshold."
                " Not plotting distribution"
            )
        )
        return

    pivot = df.pivot_table(index=groupby, columns=column, aggfunc="size", fill_value=0)

    if conditional_probability:
        pivot = pivot.divide(pivot.sum(axis=1), axis=0)  # `/` divides by axis=-1
        pivot = pivot.fillna(value=0)

    # Choose color palette
    colors = cl.scales["9"]["qual"]["Set1"]

    fig = go.Figure()
    for color_idx, (idx, row) in enumerate(pivot.iterrows()):
        if hasattr(idx, "__len__") and not isinstance(idx, str):
            group_name = "_".join([str(i) for i in idx])
        else:
            group_name = idx
        color = colors[color_idx % len(colors)]
        fig.add_trace(
            go.Bar(
                x=pivot.columns,
                y=row,
                name=group_name,
                opacity=alpha,
                marker_color=color,
            )
        )

    groupby_repr = ", ".join(groupby)
    if conditional_probability:
        yaxis_title = f"P({column} | {groupby_repr})"
    else:
        yaxis_title = f"Freq({column} | {groupby_repr})"

    fig.update_layout(
        barmode="group",
        xaxis_tickangle=xaxis_tickangle,
        xaxis_title=column,
        yaxis_title=yaxis_title,
    )
    fig.show()


def overlaid_histograms(
    df: pd.DataFrame,
    groupby: List[str],
    column: str,
    bins: Optional[int] = None,
    density: bool = True,
    alpha: float = 0.5,
):
    """Show per-group distribution histograms in a single plot overlaid over each other.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    groupby : List[str]
        List of column names to group by.
    column : str
        Name of column to analyze.
    bins : int, optional
        Number of bins in the histogram. If None, number of bin will be inferred using
        Freedman-Diaconis bin number inference.
    density : bool (default = True)
        If True, histograms will be normalized to display density.
    alpha : float
        Opacity of individual histograms.
    """
    data_min = df[column].min()
    data_max = df[column].max()
    data_range = data_max - data_min
    if bins is None:
        # Freedman-Diaconis bin number inference if bins is None
        iqr = utils.iqr(df[column])
        bin_width = 2 * iqr / (len(df[column]) ** (1 / 3))
        bins = int(np.ceil(data_range / bin_width))
        if bins > 1000:
            # Use Sturges' rule if number of bins is too large
            bins = int(np.ceil(np.log2(bins) + 1))
            bin_width = data_range / bins
    else:
        bin_width = data_range / bins
    bin_config = go.histogram.XBins(
        start=data_min,
        end=data_max,
        size=bin_width,
    )
    # Choose color palette
    colors = cl.scales["9"]["qual"]["Set1"]

    # Distribution plot
    fig = make_subplots(
        rows=2,
        cols=1,
        shared_xaxes=True,
        row_heights=[0.3, 0.7],
        vertical_spacing=0.02,
    )

    for color_idx, (name, group) in enumerate(df.groupby(groupby)):
        if hasattr(name, "__len__") and not isinstance(name, str):
            group_name = "_".join([str(i) for i in name])
        else:
            group_name = name
        color = colors[color_idx % len(colors)]
        # Add to boxplot
        fig.add_trace(
            go.Box(
                x=group[column],
                name=group_name,
                legendgroup=group_name,
                showlegend=False,
                marker_color=color,
            ),
            row=1,
            col=1,
        )
        # Add to histogram
        fig.add_trace(
            go.Histogram(
                x=group[column],
                name=group_name,
                legendgroup=group_name,
                xbins=bin_config,
                histnorm="density" if density else "",
                marker_color=color,
                opacity=alpha,
            ),
            row=2,
            col=1,
        )
    fig.update_layout(barmode="overlay")
    fig.update_xaxes(title_text=column, row=2, col=1)
    yaxis_title = "Density" if density else "Frequency"
    fig.update_yaxes(title_text=yaxis_title, row=2, col=1)
    fig.show()


def show_group_analysis(
    df: pd.DataFrame,
    groupby: Union[str, List[str]],
    columns: Optional[List[str]] = None,
    show_within_group_statistics: bool = True,
    show_group_missing_values: bool = True,
    show_distribution_plots: bool = True,
) -> None:
    """Generate group analysis for df.

    Parameters
    ----------
    df : pd.DataFrame
        Data to be analyzed.
    groupby : Union[str, List[str]]
        Name of column or list of columns names to group by.
    columns : List[str], optional
        Subset of columns to analyze. All columns except those used for grouping are used
        by default.
    show_within_group_statistics : bool (default = True)
        Whether to show per-group statistics.
    show_group_missing_values : bool (default = True)
        Whether to show per-group missing values.
    show_distribution_plots : bool (default = True)
        Whether to show per-group distribution plots.

    Raises
    ------
    ValueError
        If groupby columns are not a subset of the columns of the input DataFrame df.
    """
    if isinstance(groupby, str):
        groupby = [groupby]
    if not set(groupby) <= set(df.columns):
        raise ValueError("Grouping by a column which is not in columns of df.")
    if columns is None:
        columns = [col for col in df.columns if col not in groupby]

    if show_group_missing_values:
        display(Markdown("## Missing values for each group"))
        group_missing_values(df, groupby)
    if not show_distribution_plots and not show_within_group_statistics:
        return
    for col in columns:
        display(Markdown("---"))
        display(Markdown(f"### *{col}*"))
        datatype = infer_data_type(df[col])
        if datatype == DataType.NUMERIC:
            if show_within_group_statistics:
                within_group_stats(df, groupby, col)
            if show_distribution_plots:
                overlaid_histograms(df, groupby, col)
        else:
            group_barplot(df, groupby, col)



from typing import Any, Dict, List, Optional

import nbformat.v4 as nbfv4
import pandas as pd
import plotly.graph_objects as go
from IPython.display import Markdown, display

from edvart.data_types import is_categorical, is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class TimeSeriesLinePlot(Section):
    """Generates the time series line plot section.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    separate_plots : bool (default = False)
        Whether to plot each column in a separate plot.
        All columns are plotted in a single plot by default.
    color_col : str, optional
        Which column to use for coloring of the lines. Each segment of the line will be colored
        according to value in this column in the given time point. If this parameter is set, each
        column will be plotted in a separate plot (separate_plots param is ignored).
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        separate_plots: bool = False,
        color_col: Optional[str] = None,
    ):
        self.separate_plots = separate_plots
        self.color_col = color_col
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Time series line plot"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                "from edvart.report_sections.timeseries_analysis.time_series_line_plot"
                " import show_time_series_line_plot"
            ]
        return [
            "from IPython.display import display, Markdown",
            "import plotly.graph_objects as go",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "show_time_series_line_plot(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        if self.separate_plots:
            default_call += ", separate_plots=True"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(show_time_series_line_plot)
                + "\n\n"
                + get_code(_time_series_line_plot_colored)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates time series line plot(s) in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_time_series_line_plot(
            df=df,
            columns=self.columns,
            color_col=self.color_col,
            separate_plots=self.separate_plots,
        )


@check_index_time_ascending
def show_time_series_line_plot(
    df,
    columns: Optional[List[str]] = None,
    separate_plots: bool = False,
    color_col: Optional[str] = None,
) -> None:
    """Display time series line plot.

    Parameters
    ----------
    df: pd.DataFrame
        Data to analyze.
    columns: List[str], optional
        Columns to analyze. Only numeric columns can be analyzed.
        All numeric columns are used by default.
    separate_plots: bool (default = False)
        Whether to plot each column in a separate plot.
        All columns are plotted in a single plot by default.
    color_col: str, optional
        Name of column to use for coloring of the lines. Each segment of the line will be
        colored according to value in this column in the given time point. If this parameter is
        set, each column will be plotted in a separate plot (separate_plots param is ignored).

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    if color_col is not None:
        _time_series_line_plot_colored(df, columns=columns, color_col=color_col)
        return
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    else:
        for col in columns:
            if not is_numeric(df[col]):
                raise ValueError(
                    f"Cannot plot TimeSeriesLinePlot plot for non-numeric column {col}"
                )

    data = [go.Scatter(x=df.index, y=df[col], name=col, mode="lines") for col in columns]

    layout = dict(xaxis_rangeslider_visible=True)
    if separate_plots:
        for trace in data:
            display(Markdown(f"### {trace.name}"))
            go.Figure(data=trace, layout=layout).show()
    else:
        go.Figure(data=data, layout=layout).show()


def _time_series_line_plot_colored(df, columns=None, color_col=None):
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    else:
        for col in columns:
            if not is_numeric(df[col]):
                raise ValueError(
                    f"Cannot plot TimeSeriesLinePlot plot for non-numeric column {col}"
                )

    layout = dict(xaxis_rangeslider_visible=True)
    if not is_categorical(df[color_col]):
        raise ValueError(f"Cannot color by non-categorical column `{color_col}`")
    df_color_shifted = df[color_col].shift(-1)
    for col in columns:
        data = [
            go.Scatter(
                x=df.index,
                # GroupBy would normally be preferred, but we want a connected line
                # Therefore, we also plot a connecting line
                # to the next point where category changes
                y=df[col].mask((df[color_col] != category) & (df_color_shifted != category)),
                name=str(category),
                mode="lines",
                connectgaps=False,
            )
            for category in df[color_col].unique()
        ]
        display(Markdown(f"### {col}"))
        go.Figure(data=data, layout=layout).show()



import base64
import logging
import pickle
from abc import ABC
from copy import copy
from typing import List, Optional, Tuple, Union

import isort
import nbconvert
import nbformat as nbf
import nbformat.v4 as nbf4
import pandas as pd

from edvart.data_types import is_date
from edvart.report_sections.bivariate_analysis import BivariateAnalysis, BivariateAnalysisSubsection
from edvart.report_sections.code_string_formatting import code_dedent
from edvart.report_sections.dataset_overview import Overview, OverviewSubsection
from edvart.report_sections.group_analysis import GroupAnalysis
from edvart.report_sections.multivariate_analysis import (
    MultivariateAnalysis,
    MultivariateAnalysisSubsection,
)
from edvart.report_sections.section_base import Section, Verbosity
from edvart.report_sections.table_of_contents import TableOfContents
from edvart.report_sections.timeseries_analysis import (
    TimeseriesAnalysis,
    TimeseriesAnalysisSubsection,
)
from edvart.report_sections.univariate_analysis import UnivariateAnalysis
from edvart.utils import env_var


class ReportBase(ABC):
    """
    Abstract base class for reports.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Data from which to generate the report.
    verbosity : Verbosity (default = Verbosity.LOW)
        The default verbosity for the exported code of the entire report, by default Verbosity.LOW.
    """

    _DEFAULT_IMPORTS = {
        "import pandas as pd",
        "import os",
        "from typing import Any, Callable, Dict, List, Optional, Tuple, Union",
        "import plotly.offline as py",
        "import plotly.io as pio",
    }

    def __init__(
        self,
        dataframe: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
    ):
        self._class_logger = logging.getLogger(__name__).getChild(self.__class__.__name__)
        self.df = dataframe
        self.sections: list[Section] = []
        self.verbosity = Verbosity(verbosity)
        self._table_of_contents = None

    def show(self) -> None:
        """Renders the report in the calling notebook."""
        if self._table_of_contents is not None:
            self._table_of_contents.show(self.sections)
        for section in self.sections:
            section.show(self.df)

    def export_notebook(
        self,
        notebook_filepath: str,
        dataset_name: str = "[INSERT DATASET NAME]",
        dataset_description: str = "[INSERT DATASET DESCRIPTION]",
    ) -> None:
        """Exports the report as an .ipynb file.

        Parameters
        ----------
        notebook_filepath : str
            Filepath of the exported notebook.
        dataset_name : str (default = "[INSERT DATASET NAME]")
            Name of dataset to be used in the title of the report.
        dataset_description : str (default = "[INSERT DATASET DESCRIPTION]")
            Description of dataset to be used below the title of the report.
        """
        # Generate a notebook containing dataset name and description
        nb = self._generate_notebook(
            dataset_name=dataset_name, dataset_description=dataset_description
        )

        # Save notebook to file
        with open(notebook_filepath, "w") as notebook_file:
            nbf.write(nb, notebook_file)

    def _generate_notebook(
        self,
        dataset_name: str = "[INSERT DATASET NAME]",
        dataset_description: str = "[INSERT DATASET DESCRIPTION]",
        load_df: str = "df = ...",
        extra_imports: Optional[List[str]] = None,
        show_load_data: bool = True,
    ) -> nbf.NotebookNode:
        """Generate a notebook object for the report.

        Parameters
        ----------
        dataset_name : str (default = "[INSERT DATASET NAME]")
            Name of dataset to be used in the title of the report.
        dataset_description : str (default = "[INSERT DATASET DESCRIPTION]")
            Description of dataset to be used below the title of the report.
        load_df : str (default = 'df = ...')
            Code string for loading a dataset to variable `df`.
        extra_imports : List[str], optional
            Any additional imports to be included in imports section
            (other than imports required by report sections).
        show_load_data : bool (default = True)
            Whether to display a title for section where data is loaded.

        Returns
        -------
        nbf.NotebookNode
            Generated notebook object.
        """
        # Create notebook structure
        nb = nbf4.new_notebook()

        # Add notebook title cell in markdown
        nb["cells"].append(
            nbf4.new_markdown_cell(f"# {dataset_name} Report\n{dataset_description}")
        )

        # Add imports cell
        imports_set = copy(self._DEFAULT_IMPORTS)
        if extra_imports is not None:
            imports_set.update(extra_imports)
        for section in self.sections:
            imports_set.update(section.required_imports())

        imports_code = "\n".join(imports_set)
        imports_code = isort.code(
            imports_code, config=isort.Config(profile="black", line_length=100)
        )
        imports_code += "\n\npio.renderers.default = 'plotly_mimetype+notebook'"
        imports_code += "\npy.init_notebook_mode()"
        nb["cells"].append(nbf4.new_code_cell(imports_code))

        # Add load data cell
        if show_load_data:
            nb["cells"].append(nbf4.new_markdown_cell("## Load Data\n---"))
        nb["cells"].append(nbf4.new_code_cell(load_df))

        # Generate code for each report section
        if self._table_of_contents is not None:
            self._table_of_contents.add_cells(self.sections, nb["cells"])
        for section in self.sections:
            section.add_cells(cells=nb["cells"], df=self.df)

        return nb

    def _export_html(
        self,
        nb: nbf.NotebookNode,
        html_filepath: str,
        template_name: Optional[str] = None,
        template_filepath: Optional[str] = None,
        timeout: int = 120,
    ) -> None:
        """Export a notebook object to HTML.

        Parameters
        ----------
        nb : nbf.NotebookNode
            Notebook object to be exported to HTML.
        html_filepath : str
            File path to save exported HTML to.
        template_name: str, optional
            Name of the template to use for exporting the notebook.

            The template must be found in a Jupyter path
            (see
            https://nbconvert.readthedocs.io/en/latest/customizing.html#where-are-nbconvert-templates-installed
            ).
            The default location is `$HOME/.local/share/jupyter/nbconvert/templates`
        template_filepath : str, optional
            Path to template file to use for exporting the notebook to HTML.
            Overrides the parameter `template_name` if specified.
        timeout: int (default = 120)
            Maximum number of seconds to wait for a cell to finish execution.
        """
        # Execute notebook to produce output of cells
        html_exp_kwargs = dict(
            preprocessors=[nbconvert.preprocessors.ExecutePreprocessor(timeout=timeout)]
        )
        if template_name is not None:
            html_exp_kwargs["template_name"] = template_name
        elif template_filepath is not None:
            html_exp_kwargs["template_file"] = template_filepath
        else:
            html_exp_kwargs["exclude_input"] = True

        html_exporter = nbconvert.HTMLExporter(**html_exp_kwargs)

        # Workaround for a warning from `nbconvert` regarding debugging
        # and frozen modules. We are not debugging, so we can safely ignore it.
        with env_var("PYDEVD_DISABLE_FILE_VALIDATION", "1"):
            html = html_exporter.from_notebook_node(nb)[0]

        # Save HTML to file
        with open(html_filepath, "w") as html_file:
            print(html, file=html_file)

    def export_html(
        self,
        html_filepath: str,
        template_name: Optional[str] = None,
        template_filepath: Optional[str] = None,
        dataset_name: str = "[INSERT DATASET NAME]",
        dataset_description: str = "[INSERT DATASET DESCRIPTION]",
        timeout: int = 120,
    ) -> None:
        """Generate HTML report for an already-loaded DataFrame.

        Parameters
        ----------
        html_filepath: str
            File path to save exported HTML report to.
        template_name: str, optional
            Path to template file to use for exporting the notebook to HTML.

            The template must be found in a Jupyter path
            (see
            https://nbconvert.readthedocs.io/en/latest/customizing.html#where-are-nbconvert-templates-installed
            ).
            The default location is `$HOME/.local/share/jupyter/nbconvert/templates`
        template_filepath: str, optional
            Template to use when exporting the HTML report.
        dataset_name : str (default = "[INSERT DATASET NAME]")
            Name of dataset to be used in the title of the report.
        dataset_description : str (default = "[INSERT DATASET DESCRIPTION]")
            Description of dataset to be used below the title of the report.
        timeout: int (default = 120)
            Maximum number of seconds to wait for a cell to finish execution.
        """
        self._class_logger.info(
            "Exporting report to HTML requires generating and executing a notebook, "
            "which may take some time."
        )

        # Pickle self (the whole report) and encode it into a base64 string
        buffer = pickle.dumps(self, fix_imports=False)
        buffer_base64 = base64.b85encode(buffer)

        # Prepare code to inject into notebook,
        # which decodes data from base64 (injected as string literal)
        # and unpickles the the whole report object from the decoded binary data
        unpickle_report = code_dedent(
            f"""
                data = {buffer_base64}
                report = pickle.loads(base64.b85decode(data), fix_imports=False)
            """
        )

        # Create notebook object
        nb = nbf4.new_notebook()

        # Add title and description
        nb["cells"].append(
            nbf4.new_markdown_cell(f"# {dataset_name} Report\n{dataset_description}")
        )

        # Add code cells to generated notebook
        for code_string in (
            "import base64",
            "import pickle",
            unpickle_report,
            "import plotly.io as pio",
            "pio.renderers.default = 'notebook'",
            "report.show()",
        ):
            nb["cells"].append(nbf4.new_code_cell(code_string))

        self._export_html(
            nb=nb,
            html_filepath=html_filepath,
            template_filepath=template_filepath,
            template_name=template_name,
            timeout=timeout,
        )

    def _dev_export_notebook(self, notebook_filepath: str) -> None:
        """Same as export_notebook, but conveniently adds dataset into exported notebook for
        development purposes.
        """
        nb = self._generate_notebook(
            extra_imports=["import edvart"],
            load_df="df = edvart.example_datasets.dataset_titanic()",
        )

        # Save notebook to file
        with open(notebook_filepath, "w") as notebook_file:
            nbf.write(nb, notebook_file)

    def add_section(
        self,
        section: Section,
    ) -> "ReportBase":
        """
        Add a section to the report. See `edvart.report_sections` for available sections.

        Parameters
        ----------
        section : Section
            Section to add to the report.

        Returns
        -------
        ReportBase
            Returns self.
        """
        self.sections.append(section)

        return self

    def add_overview(
        self,
        columns: Optional[List[str]] = None,
        subsections: Optional[List[OverviewSubsection]] = None,
        verbosity: Optional[Verbosity] = None,
        verbosity_quick_info: Optional[Verbosity] = None,
        verbosity_data_types: Optional[Verbosity] = None,
        verbosity_data_preview: Optional[Verbosity] = None,
        verbosity_missing_values: Optional[Verbosity] = None,
        verbosity_rows_with_missing_value: Optional[Verbosity] = None,
        verbosity_constant_occurrence: Optional[Verbosity] = None,
        verbosity_duplicate_rows: Optional[Verbosity] = None,
    ) -> "ReportBase":
        """
        Adds a dataset overview section to the report.

        Parameters
        ----------
        columns : List[str], optional
            Columns which to include in the overview section.
            If None, all columns are used.
        subsections : List[Overview.OverviewSubsection], optional
            List of sub-sections to include into the Overview section.
            If None, all subsections are added.
        verbosity : Verbosity, optional
            Generated code verbosity global to the Overview sections.
            If subsection verbosities are None, then they will be overridden by this parameter.
        verbosity_quick_info : Verbosity, optional
            Quick info sub-section code verbosity.
        verbosity_data_types : Verbosity, optional
            Data types sub-section code verbosity.
        verbosity_data_preview : Verbosity, optional
            Data preview sub-section code verbosity.
        verbosity_missing_values : Verbosity, optional
            Missing values sub-section code verbosity.
        verbosity_rows_with_missing_value : Verbosity, optional
            Rows with missing value sub-section code verbosity.
        verbosity_constant_occurrence : Verbosity, optional
            Constant values occurrence sub-section code verbosity.
        verbosity_duplicate_rows : Verbosity, optional
            Duplicate rows sub-section code verbosity.

        """
        # Construct and store overview configuration
        self.sections.append(
            Overview(
                subsections=subsections,
                verbosity=verbosity or self.verbosity,
                columns=columns,
                verbosity_quick_info=verbosity_quick_info,
                verbosity_data_types=verbosity_data_types,
                verbosity_data_preview=verbosity_data_preview,
                verbosity_missing_values=verbosity_missing_values,
                verbosity_rows_with_missing_value=verbosity_rows_with_missing_value,
                verbosity_constant_occurrence=verbosity_constant_occurrence,
                verbosity_duplicate_rows=verbosity_duplicate_rows,
            )
        )
        return self

    def add_univariate_analysis(
        self,
        columns: Optional[List[str]] = None,
        verbosity: Optional[Verbosity] = None,
    ) -> "ReportBase":
        """Adds univariate section to the report.

        Parameters
        ----------
        columns : List[str], optional
            Columns which to analyze.
            If None, all columns are used.
        verbosity : Verbosity
            The verbosity of the code generated in the exported notebook.
        """
        self.sections.append(
            UnivariateAnalysis(
                verbosity=verbosity or self.verbosity,
                columns=columns,
            )
        )
        return self

    def add_bivariate_analysis(
        self,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
        subsections: Optional[List[BivariateAnalysisSubsection]] = None,
        verbosity: Optional[Verbosity] = None,
        verbosity_correlations: Optional[Verbosity] = None,
        verbosity_pairplot: Optional[Verbosity] = None,
        verbosity_contingency_table: Optional[Verbosity] = None,
        color_col: Optional[str] = None,
    ) -> "ReportBase":
        """Adds bivariate analysis section to the report.

        Parameters
        ----------
        columns : List[str], optional
            Columns which to analyze.
            If None, all columns are used.
        columns_x : List[str], optional
            If specified, correlations and pairplots are performed on the cartesian product of
            `columns_x` and `columns_y`.
            If `columns_x` is specified, then `columns_y` must also be specified.
        columns_y : List[str], optional
            If specified, correlations and pairplots are performed on the cartesian product of
            `columns_x` and `columns_y`.
            If `columns_y` is specified, then `columns_x` must also be specified.
        columns_pairs : List[str], optional
            List of columns pairs on which to perform bivariate analysis.
            Used primarily in contingency tables.
            If specified, `columns`, `columns_x` and `columns_y` are ignored in contingency tables.
            Ignored in pairplots and correlations unless `columns_pairs` is specified and none of
            `columns`, `columns_x`, `columns_y` is specified. In that case, the first elements
            of each pair are treated as `columns_x` and the second elements as `columns_y` in
            pairplots and correlations.
        subsections : List[BivariateAnalysisSubsection], optional
            List of sub-sections to include into the BivariateAnalysis section.
            If None, all subsections are added.
        verbosity : Verbosity, optional
            The verbosity of the code generated in the exported notebook.
        verbosity_correlations : Verbosity, optional
            Correlation plots subsection code verbosity.
        verbosity_pairplot : Verbosity, optional
            Pairplot subsection code verbosity.
        verbosity_contingency_table : Verbosity, optional
            Contingency table code verbosity.
        color_col : str, optional
            Name of column according to use for coloring of the multivariate analysis subsections.
            Coloring is currently supported in pairplot.
        """
        self.sections.append(
            BivariateAnalysis(
                subsections=subsections,
                verbosity=verbosity or self.verbosity,
                columns=columns,
                columns_x=columns_x,
                columns_y=columns_y,
                columns_pairs=columns_pairs,
                verbosity_correlations=verbosity_correlations,
                verbosity_pairplot=verbosity_pairplot,
                verbosity_contingency_table=verbosity_contingency_table,
                color_col=color_col,
            )
        )

        return self

    def add_multivariate_analysis(
        self,
        columns: Optional[List[str]] = None,
        subsections: Optional[List[MultivariateAnalysisSubsection]] = None,
        verbosity: Optional[Verbosity] = None,
        verbosity_pca: Optional[Verbosity] = None,
        verbosity_umap: Optional[Verbosity] = None,
        verbosity_parallel_coordinates: Optional[Verbosity] = None,
        verbosity_parallel_categories: Optional[Verbosity] = None,
        color_col: Optional[str] = None,
    ) -> "ReportBase":
        """Add multivariate analysis section to the report.

        Parameters
        ----------
        columns : List[str], optional
            Columns which to analyze.
            If None, all columns are used.
        subsections : List[MultivariateAnalysisSubsection], optional
            List of sub-sections to include into the BivariateAnalysis section.
            If None, all subsections are added.
        verbosity : Verbosity, optional
            The verbosity of the code generated in the exported notebook.
        verbosity_pca : Verbosity, optional
            Principal component analysis subsection code verbosity.
        verbosity_umap : Verbosity, optional
            UMAP subsection code verbosity.
        verbosity_parallel_coordinates: Verbosity, optional
            Parallel coordinates subsection code verbosity.
        verbosity_parallel_categories: Verbosity, optional
            Parallel categories subsection code verbosity.
        color_col : str, optional
            Name of column to use for coloring of the multivariate analysis subsections.
            The exact method of coloring depends on each particular subsection.
        """
        self.sections.append(
            MultivariateAnalysis(
                subsections=subsections,
                verbosity=verbosity or self.verbosity,
                columns=columns,
                verbosity_pca=verbosity_pca,
                verbosity_umap=verbosity_umap,
                verbosity_parallel_coordinates=verbosity_parallel_coordinates,
                verbosity_parallel_categories=verbosity_parallel_categories,
                color_col=color_col,
            )
        )

        return self

    def add_group_analysis(
        self,
        groupby: Union[str, List[str]],
        columns: Optional[List[str]] = None,
        verbosity: Optional[Verbosity] = None,
        show_within_group_statistics: bool = True,
        show_group_missing_values: bool = True,
        show_group_distribution_plots: bool = True,
    ) -> "ReportBase":
        """Add group analysis section to the report.

        Parameters
        ----------
        groupby: Union[str, List[str]]
            Column or list of columns to group by in group analysis.
        columns : List[str], optional
            Columns which to analyze.
            If None, all columns are used.
        verbosity : Verbosity, optional
            The verbosity of the code generated in the exported notebook.
        show_within_group_statistics : bool (default = True)
            Whether to show per-group statistics.
        show_group_missing_values : bool (default = True)
            Whether to show per-group missing values.
        show_group_distribution_plots : bool (default = True)
            Whether to show per-group distribution plots.
        """
        self.sections.append(
            GroupAnalysis(
                groupby=groupby,
                verbosity=verbosity or self.verbosity,
                columns=columns,
                show_within_group_statistics=show_within_group_statistics,
                show_group_missing_values=show_group_missing_values,
                show_group_distribution_plots=show_group_distribution_plots,
            )
        )

        return self

    def add_table_of_contents(self, include_subsections: bool = True) -> "ReportBase":
        """Adds table of contents section to the report.

        Parameters
        ----------
        include_subsections: bool
            A boolean controlling whether the subsections should be included in the table of
            contents. However, they won't be included in an exported notebook created by report's
            export_notebook function.
        """
        self._table_of_contents = TableOfContents(include_subsections)
        return self


class Report(ReportBase):
    """
    A report for tabular datasets. Contains no sections by default.

    See `DefaultReport` for a report with default sections.
    See methods `add_*` for adding sections to the report.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Data from which to generate the report.
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the exported code of the entire report.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
    ):
        super().__init__(dataframe=dataframe, verbosity=verbosity)


class DefaultReport(Report):
    """A report for tabular data containing default sections.

    The report contains the following sections:
    - dataset overview
    - univariate analysis
    - bivariate analysis
    - multivariate analysis
    - group analysis (if `groupby` is specified)

    Parameters
    ----------
    dataframe : pd.DataFrame
        Data from which to generate the report.
    verbosity : Verbosity (default = Verbosity.LOW)
        The default verbosity for the exported code of the entire report.
    verbosity_overview : Verbosity, optional
        Verbosity of the overview section
    verbosity_univariate_analysis : Verbosity, optional
        Verbosity of the univariate analysis section
    verbosity_bivariate_analysis : Verbosity, optional
        Verbosity of the bivariate analysis section.
    verbosity_multivariate_analysis: Verbosity, optional
        Verbosity of the multivariate analysis section
    columns_overview : List[str], optional
        Subset of columns to use in overview section
    columns_univariate_analysis : List[str], optional
        Subset of columns to use in univariate analysis section
    columns_bivariate_analysis: List[str], optional
        Subset of columns to use in bivariate analysis section
    columns_multivariate_analysis: List[str], optional
        Subset of columns to use in multivariate analysis section
    columns_group_analysis: List[str], optional
        Subset of columns to use in group analysis section
    groupby: Union[str, List[str]], optional
        Column or list of columns to group by in group analysis. If None, group analysis will not be
        included by default. It can still be added later using `add_group_analysis`.
        If a single column is specified, it will be used to color points in multivariate analysis.
        Default: None.

    """

    # pylint:disable=too-many-locals
    def __init__(
        self,
        dataframe: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
        verbosity_overview: Optional[Verbosity] = None,
        verbosity_univariate_analysis: Optional[Verbosity] = None,
        verbosity_bivariate_analysis: Optional[Verbosity] = None,
        verbosity_multivariate_analysis: Optional[Verbosity] = None,
        verbosity_group_analysis: Optional[Verbosity] = None,
        columns_overview: Optional[List[str]] = None,
        columns_univariate_analysis: Optional[List[str]] = None,
        columns_bivariate_analysis: Optional[List[str]] = None,
        columns_multivariate_analysis: Optional[List[str]] = None,
        columns_group_analysis: Optional[List[str]] = None,
        groupby: Union[str, List[str]] = None,
    ):
        super().__init__(dataframe, verbosity)

        # If section verbosities are not set, default to the global verbosity
        if verbosity_overview is None:
            verbosity_overview = verbosity
        if verbosity_univariate_analysis is None:
            verbosity_univariate_analysis = verbosity
        if verbosity_bivariate_analysis is None:
            verbosity_bivariate_analysis = verbosity
        if verbosity_multivariate_analysis is None:
            verbosity_multivariate_analysis = verbosity

        # Add default sections
        self.add_table_of_contents(include_subsections=True)
        self.add_overview(verbosity=verbosity_overview, columns=columns_overview)
        self.add_univariate_analysis(
            verbosity=verbosity_univariate_analysis,
            columns=columns_univariate_analysis,
        )
        if isinstance(groupby, str):
            color_col = groupby
        elif hasattr(groupby, "__len__") and len(groupby) == 1:
            color_col = groupby[0]
        else:
            color_col = None
        self.add_bivariate_analysis(
            verbosity=verbosity_bivariate_analysis,
            columns=columns_bivariate_analysis,
            color_col=color_col,
        )
        self.add_multivariate_analysis(
            verbosity=verbosity_multivariate_analysis,
            columns=columns_multivariate_analysis,
            color_col=color_col,
        )
        if groupby is not None:
            self.add_group_analysis(
                groupby=groupby,
                columns=columns_group_analysis,
                verbosity=verbosity_group_analysis,
            )


class TimeseriesReport(ReportBase):
    """
    A report for time-series data. Contains no sections by default.

    See `DefaultTimeseriesReport` for a time-series report with default sections.
    See methods `add_*` for adding sections to the report.

    Raises
    ------
    ValueError
        If the input dataframe is not indexed by time.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
    ):
        super().__init__(dataframe, verbosity)
        if not is_date(dataframe.index):
            raise ValueError(
                "Input dataframe needs to be indexed by time."
                "Please reindex your data to be indexed by either a DatetimeIndex or a PeriodIndex."
                "See `edvart.utils.reindex_to_datetime`, `edvart.utils.reindex_to_period`."
            )
        dataframe = dataframe.copy()
        if isinstance(dataframe.index.dtype, pd.PeriodDtype):
            dataframe.index = pd.PeriodIndex(dataframe.index)
        else:
            dataframe.index = pd.DatetimeIndex(dataframe.index)

    def add_timeseries_analysis(
        self,
        columns: Optional[List[str]] = None,
        subsections: Optional[List[TimeseriesAnalysisSubsection]] = None,
        verbosity: Optional[Verbosity] = None,
        verbosity_time_series_line_plot: Optional[Verbosity] = None,
        verbosity_rolling_statistics: Optional[Verbosity] = None,
        verbosity_boxplots_over_time: Optional[Verbosity] = None,
        verbosity_seasonal_decomposition: Optional[Verbosity] = None,
        verbosity_autocorrelation: Optional[Verbosity] = None,
        verbosity_stationarity_tests: Optional[Verbosity] = None,
        verbosity_fourier_transform: Optional[Verbosity] = None,
        verbosity_short_time_ft: Optional[Verbosity] = None,
        sampling_rate: Optional[Verbosity] = None,
        stft_window_size: Optional[Verbosity] = None,
    ) -> "TimeseriesReport":
        """Add timeseries analysis section to the report.

        Parameters
        ----------
        columns : List[str], optional
            Columns which to analyze.
            If None, all columns are used.
        subsections : List[TimeseriesAnalysis.TimeseriesAnalysisSubsection], optional
            List of sub-sections to include into the BivariateAnalysis section.
            If None, all subsections are added.
        verbosity : Verbosity, optional
            The verbosity of the code generated in the exported notebook.
        verbosity_time_series_line_plot : Verbosity, optional
            Time series line plot subsection code verbosity.
        verbosity_rolling_statistics : Verbosity, optional
            Rolling statistics interactive plot subsection code verbosity.
        verbosity_boxplots_over_time : Verbosity, optional
            Boxplots grouped over time intervals plot subsection code verbosity.
        verbosity_seasonal_decomposition : Verbosity, optional
            Decomposition into trend, seasonal and residual components code verbosity.
        verbosity_autocorrelation : Verbosity, optional
            Autocorrelation and partial autocorrelation vs. lag code verbosity.
        verbosity_stationarity_tests : Verbosity, optional
            Stationarity tests code verbosity.
        verbosity_fourier_transform: Verbosity, optional
            Fourier transform and short-time Fourier transform code verbosity.
        verbosity_short_time_ft: Verbosity, optional
            Short-time Fourier transform transform spectrogram code verbosity.
        sampling_rate: Verbosity, optional
            Sampling rate for Fourier transform and Short-time Fourier transform subsections.
            Needs to be set in order for these two subs to be included.
        stft_window_size : Verbosity, optional
            Window size for Short-time Fourier transform. Needs to be set in order for the STFT
            subsection to be included.
        """
        self.sections.append(
            TimeseriesAnalysis(
                subsections=subsections,
                verbosity=verbosity or self.verbosity,
                columns=columns,
                verbosity_time_series_line_plot=verbosity_time_series_line_plot,
                verbosity_rolling_statistics=verbosity_rolling_statistics,
                verbosity_boxplots_over_time=verbosity_boxplots_over_time,
                verbosity_seasonal_decomposition=verbosity_seasonal_decomposition,
                verbosity_autocorrelation=verbosity_autocorrelation,
                verbosity_stationarity_tests=verbosity_stationarity_tests,
                verbosity_fourier_transform=verbosity_fourier_transform,
                verbosity_short_time_ft=verbosity_short_time_ft,
                sampling_rate=sampling_rate,
                stft_window_size=stft_window_size,
            )
        )

        return self


class DefaultTimeseriesReport(TimeseriesReport):
    """A default report for time series data.

    The report contains the following sections:
    - dataset overview
    - univariate analysis
    - timeseries analysis

    Parameters
    ----------
    dataframe : pd.DataFrame
        Data from which to generate the report. Data needs to be indexed by time: pd.DateTimeIndex
        or pd.PeriodIndex.
        The data is assumed to be sorted according to the time index in ascending order.
    verbosity : Verbosity (default = Verbosity.LOW)
        The default verbosity for the exported code of the entire report.
    verbosity_overview : Verbosity, optional
        Verbosity of the overview section
    verbosity_univariate_analysis : Verbosity, optional
        Verbosity of the univariate analysis section
    verbosity_timeseries_analysis : Verbosity, optional
        Verbosity of the timeseries analysis section
    columns_overview : List[str], optional
        Subset of columns to use in overview section
    columns_univariate_analysis : List[str], optional
        Subset of columns to use in univariate analysis section
    columns_timeseries_analysis : List[str], optional
        Subset of columns to use in timeseries analysis section
    sampling_rate : int, optional
        Sampling rate for Fourier transform and Short-time Fourier transform subsections. Determines
        frequency unit for analysis of frequencies, for example with monthly data and sampling rate
        12, yearly frequency spectrum is produced.
        If not set, these two sections will not be included.
    stft_window_size : int, optional
        Windows size for short-time Fourier transform subsection. If not set, STFT will be excluded.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        verbosity: Verbosity = Verbosity.LOW,
        verbosity_overview: Optional[Verbosity] = None,
        verbosity_univariate_analysis: Optional[Verbosity] = None,
        verbosity_timeseries_analysis: Optional[Verbosity] = None,
        columns_overview: Optional[List[str]] = None,
        columns_univariate_analysis: Optional[List[str]] = None,
        columns_timeseries_analysis: Optional[List[str]] = None,
        sampling_rate: Optional[Verbosity] = None,
        stft_window_size: Optional[Verbosity] = None,
    ):
        super().__init__(dataframe, verbosity)

        if verbosity_overview is None:
            verbosity_overview = verbosity
        if verbosity_univariate_analysis is None:
            verbosity_univariate_analysis = verbosity
        if verbosity_timeseries_analysis is None:
            verbosity_timeseries_analysis = verbosity
        self.add_table_of_contents(include_subsections=True)
        self.add_overview(verbosity=verbosity_overview, columns=columns_overview)
        self.add_univariate_analysis(
            verbosity=verbosity_univariate_analysis,
            columns=columns_univariate_analysis,
        )
        self.add_timeseries_analysis(
            verbosity=verbosity_timeseries_analysis,
            columns=columns_timeseries_analysis,
            sampling_rate=sampling_rate,
            stft_window_size=stft_window_size,
        )



from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import sklearn.decomposition
from IPython.display import Markdown, display
from sklearn.preprocessing import StandardScaler

from edvart.data_types import DataType, infer_data_type, is_boolean, is_categorical, is_numeric
from edvart.plots import scatter_plot_2d
from edvart.report_sections.code_string_formatting import code_dedent, get_code
from edvart.report_sections.section_base import ReportSection, Section, Verbosity
from edvart.utils import (
    get_default_discrete_colorscale,
    hsl_wheel_colorscale,
    make_discrete_colorscale,
    select_numeric_columns,
)

try:
    from edvart.report_sections.umap import UMAP  # pylint: disable=cyclic-import
except ImportError:
    UMAP_AVAILABLE = False
else:
    UMAP_AVAILABLE = True


# pylint:disable=invalid-name
class MultivariateAnalysisSubsection(IntEnum):
    """Enum of all implemented multivariate analysis subsections."""

    PCA = 0
    if UMAP_AVAILABLE:
        UMAP = 1
    ParallelCoordinates = 2
    ParallelCategories = 3

    def __str__(self):
        return self.name


class MultivariateAnalysis(ReportSection):
    """Generates the Multivariate analysis section of the report.

    Contains an enum MultivariateAnalysisSubsection of possible subsections.

    Parameters
    ----------
    subsections : List[MultivariateAnalysisSubsection], optional
        List of subsections to include.
        All subsection in MultivariateAnalysisSubsection are included by default.
    verbosity : Verbosity
        Generated code verbosity global to the Multivariate sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do multivariate analysis.
        All columns of df will be used by default.
    verbosity_pca : Verbosity, optional
        Principal component analysis subsection code verbosity.
    verbosity_umap : Verbosity, optional
        UMAP subsection code verbosity.
    verbosity_parallel_coordinates : Verbosity, optional
        Parallel coordinates subsection code verbosity.
    verbosity_parallel_categories : Verbosity, optional
        Parallel categories subsection code verbosity.
    color_col : str, optional
        Name of the column according to which to color points in the sections.
        Both numerical and categorical columns are supported.
    """

    # By default use all subsections
    _DEFAULT_SUBSECTIONS_TO_SHOW = list(MultivariateAnalysisSubsection)

    def __init__(
        self,
        subsections: Optional[List[MultivariateAnalysisSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        verbosity_pca: Optional[Verbosity] = None,
        verbosity_umap: Optional[Verbosity] = None,
        verbosity_parallel_coordinates: Optional[Verbosity] = None,
        verbosity_parallel_categories: Optional[Verbosity] = None,
        color_col: Optional[str] = None,
    ):
        verbosity_pca = verbosity_pca or verbosity
        verbosity_umap = verbosity_umap or verbosity
        verbosity_parallel_coordinates = (
            verbosity_parallel_coordinates
            if verbosity_parallel_coordinates is not None
            else verbosity
        )
        verbosity_parallel_categories = (
            verbosity_parallel_categories
            if verbosity_parallel_categories is not None
            else verbosity
        )

        self.subsection_verbosities = {
            MultivariateAnalysisSubsection.PCA: verbosity_pca,
            MultivariateAnalysisSubsection.ParallelCoordinates: verbosity_parallel_coordinates,
            MultivariateAnalysisSubsection.ParallelCategories: verbosity_parallel_categories,
        }
        if UMAP_AVAILABLE:
            self.subsection_verbosities[MultivariateAnalysisSubsection.UMAP] = verbosity_umap

        if subsections is None:
            self.subsections_to_show = self._DEFAULT_SUBSECTIONS_TO_SHOW
        else:
            self.subsections_to_show = subsections

        self.subsections_to_show_with_low_verbosity = [
            sub
            for sub in self.subsections_to_show
            if self.subsection_verbosities[sub] == Verbosity.LOW
        ]

        enum_to_implementation = {
            MultivariateAnalysisSubsection.PCA: PCA(verbosity_pca, columns, color_col=color_col),
            MultivariateAnalysisSubsection.ParallelCoordinates: ParallelCoordinates(
                verbosity_parallel_coordinates, columns, color_col=color_col
            ),
            MultivariateAnalysisSubsection.ParallelCategories: ParallelCategories(
                verbosity_parallel_categories, columns, color_col=color_col
            ),
        }
        if UMAP_AVAILABLE:
            enum_to_implementation[MultivariateAnalysisSubsection.UMAP] = UMAP(
                verbosity_umap, columns, color_col=color_col
            )

        subsections_implementations = [
            enum_to_implementation[sub] for sub in self.subsections_to_show
        ]

        self.color_col = color_col
        super().__init__(subsections_implementations, verbosity, columns)

    @property
    def name(self) -> str:
        return "Multivariate analysis"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity == Verbosity.LOW:
            imports = {
                "from edvart.report_sections.multivariate_analysis import"
                " show_multivariate_analysis"
            }
            if self.subsections_to_show_with_low_verbosity != self._DEFAULT_SUBSECTIONS_TO_SHOW:
                imports.add(
                    "from edvart.report_sections.multivariate_analysis import"
                    " MultivariateAnalysisSubsection"
                )
            for subsec in self.subsections:
                if subsec.verbosity > Verbosity.LOW:
                    imports.update(subsec.required_imports())

            return list(imports)
        return super().required_imports()

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)
        if self.verbosity == Verbosity.LOW:
            code = "show_multivariate_analysis(df=df"
            if self.subsections_to_show_with_low_verbosity != self._DEFAULT_SUBSECTIONS_TO_SHOW:
                arg_subsections_names = [
                    f"MultivariateAnalysisSubsection.{str(sub)}"
                    for sub in self.subsections_to_show_with_low_verbosity
                ]
                code += f", subsections={arg_subsections_names}".replace("'", "")
            if self.columns is not None:
                code += f", columns={self.columns}"
            if self.color_col is not None:
                code += f", color_col='{self.color_col}'"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))
            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    sub.add_cells(cells=cells, df=df)
        else:
            super().add_cells(cells=cells, df=df)

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)


def show_multivariate_analysis(
    df: pd.DataFrame,
    subsections: Optional[List[MultivariateAnalysisSubsection]] = None,
    columns: Optional[List[str]] = None,
    color_col: Optional[str] = None,
) -> None:
    """Generates multivariate analysis for df.

    Parameters
    ----------
    df : pd.DataFrame
        Data to be analyzed
    subsections : List[MultivariateAnalysisSubsection], optional
        Subsections to include in the analysis. All subsections are included by default.
    columns : List[str], optional
        Subset of columns of df to consider in multivariate analysis.
        All numeric columns are used by default.
    color_col : str, optional
        Name of the column according to which to color points in the sections.
        Both numeric and categorical columns are supported.
    """
    if columns is not None:
        df = df[columns]

    multivariate_analysis = MultivariateAnalysis(
        subsections=subsections,
        verbosity=Verbosity.LOW,
        columns=columns,
        color_col=color_col,
    )

    for sub in multivariate_analysis.subsections:
        sub.show(df)


class PCA(Section):
    """Generates the Principal component analysis subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to perform PCA. Only numeric columns can be used.
        All numeric columns of df are used by default.
    color_col : str, optional
        Name of column according to values of which to color points in the first vs second component
        plot. Can be both numeric and categorical. By default, all points have the same color.
    standardize : bool (default = True)
        Whether to standardize the data to zero mean and unit variance before applying PCA.
    interactive : bool (default = True)
        Whether to plot the first vs second principal component as an interactive plot.
        The interactive plot also shows labels for each point on hover.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
        standardize: bool = True,
        interactive: bool = True,
    ):
        self.color_col = color_col
        self.interactive = interactive
        self.standardize = standardize
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Principal Component Analysis"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                code_dedent(
                    """from edvart.report_sections.multivariate_analysis import (
                        pca_explained_variance, pca_first_vs_second
                    )"""
                )
            ]
        return [
            "from edvart.plots import scatter_plot_2d",
            "import matplotlib.pyplot as plt",
            "import numpy as np",
            "import sklearn.decomposition",
            "from sklearn.preprocessing import StandardScaler",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        first_vs_second_header = nbfv4.new_markdown_cell("### First vs second principal component")
        cells.append(first_vs_second_header)
        explained_variance_header = nbfv4.new_markdown_cell("### Explained variance ratio")

        first_vs_second_call = "pca_first_vs_second(df=df"
        if not self.interactive:
            first_vs_second_call += ", interactive=False"
        explained_variance_call = "pca_explained_variance(df=df"
        if self.columns is not None:
            first_vs_second_call += f", columns={self.columns}"
            explained_variance_call += f", columns={self.columns}"
        if not self.standardize:
            explained_variance_call += ", standardize=False"
            first_vs_second_call += ", standardize=False"
        if not self.interactive:
            first_vs_second_call += ", interactive=False"
        if self.color_col is not None:
            first_vs_second_call += f", color_col='{self.color_col}'"
        first_vs_second_call += ")"
        explained_variance_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            cells.append(nbfv4.new_code_cell(first_vs_second_call))
            cells.append(explained_variance_header)
            cells.append(nbfv4.new_code_cell(explained_variance_call))
        else:
            select_columns_code = get_code(select_numeric_columns)
            cells.append(nbfv4.new_code_cell(select_columns_code))

            first_vs_second_code = get_code(pca_first_vs_second) + "\n\n" + first_vs_second_call
            cells.append(nbfv4.new_code_cell(first_vs_second_code))

            cells.append(explained_variance_header)
            explained_variance_code = (
                get_code(pca_explained_variance) + "\n\n" + explained_variance_call
            )
            cells.append(nbfv4.new_code_cell(explained_variance_code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the PCA section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        display(Markdown("### First vs second principal component"))
        pca_first_vs_second(
            df=df,
            columns=self.columns,
            color_col=self.color_col,
            interactive=self.interactive,
            standardize=self.standardize,
        )
        display(Markdown("### Explained variance ratio"))
        pca_explained_variance(df=df, columns=self.columns, standardize=self.standardize)


def pca_first_vs_second(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    color_col: Optional[str] = None,
    interactive: bool = True,
    standardize: bool = True,
    figsize: Tuple[float, float] = (12, 12),
    opacity: float = 0.8,
) -> None:
    """Plot a 2D scatter of first vs second PCA components.

    Parameters
    ----------
    df : pd.DataFrame
        Data to perform PCA on.
    columns : List[str], optional
        Which columns to perform PCA on. All columns will be used by default.
    color_col : str, optional
        Name of column according to values of which to color points in the plot.
        Can be both numeric and categorical. By default, all points have the same color.
    interactive : bool (default = True)
        Whether to show an interactive plot.
    standardize : bool (default = True)
        Whether to standardize the data to zero mean and unit variance before applying PCA.
    figsize : Tuple[float, float] (default = (12, 12))
        Size of the plot.
    opacity : float (default = 0.8)
        Opacity of the points in the plot. Higher means more opaque (less transparent).
    """
    columns = select_numeric_columns(df, columns)
    df = df.dropna(subset=columns)

    pca = sklearn.decomposition.PCA(n_components=2)

    if standardize:
        scaler = StandardScaler()
        data_scaled = scaler.fit_transform(df[columns])
        pca_components = pca.fit_transform(data_scaled)
    else:
        pca_components = pca.fit_transform(df[columns])

    scatter_plot_2d(
        df=df,
        x=pca_components[:, 0],
        y=pca_components[:, 1],
        color_col=color_col,
        interactive=interactive,
        figsize=figsize,
        opacity=opacity,
        xlabel="First principal component",
        ylabel="Second principal component",
        show_xticks=True,
        show_yticks=True,
        show_zerolines=True,
        equal_scale_axes=True,
    )

    print(f"Explained variance ratio: {pca.explained_variance_ratio_[:2].sum() * 100 :.2f}%")


def pca_explained_variance(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    standardize: bool = True,
    show_grid: bool = True,
    figsize: Tuple[float, float] = (10, 7),
) -> None:
    """Plot a plot of variance explained by each principal component.

    Parameters
    ----------
    df : pd.DataFrame
        Data on which to perform PCA.
    columns : List[str], optional
        Which columns to perform PCA on. All columns will be used by default.
    standardize : bool (default = True)
        Whether to standardize the data zero mean and unit variance before applying PCA.
    show_grid : bool (default = True)
        Whether to show a grid in the plot.
    figsize : Tuple[float, float] (default = (10, 7))
        Size of the plot.
    """
    columns = select_numeric_columns(df, columns)
    df = df.dropna(subset=columns)

    pca = sklearn.decomposition.PCA()

    if standardize:
        scaler = StandardScaler()
        data_scaled = scaler.fit_transform(df[columns])
        pca.fit(data_scaled)
    else:
        pca.fit(df[columns])

    fig = plt.figure(figsize=figsize)
    plt.plot(pca.explained_variance_ratio_, figure=fig)
    plt.plot(np.cumsum(pca.explained_variance_ratio_), figure=fig)

    plt.legend(["Individual component", "Cumulative"])
    plt.xlabel("Principal component #")
    plt.ylabel("Explained variance ratio")
    plt.xticks(
        ticks=range(len(pca.explained_variance_ratio_)),
        labels=range(1, (len(pca.explained_variance_ratio_) + 1)),
    )
    if show_grid:
        plt.grid()
    plt.show()


class ParallelCoordinates(Section):
    """Generates the Parallel coordinates subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns for which to generate parallel coordinates. All columns which are either numeric or
        categorical with at most `nunique_max` unique values are used by default.
    color_col : str, optional
        Name of column determining color of the coordinate lines.
        Both numeric and categorical columns are supported.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
    ):
        self.color_col = color_col
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Parallel coordinates"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.multivariate_analysis import parallel_coordinates"]
        return [
            "from edvart.utils import ("
            "    get_default_discrete_colorscale, make_discrete_colorscale, hsl_wheel_colorscale"
            ")",
            "from typing import Iterable",
            "import plotly",
            "import plotly.graph_objects as go",
            "from edvart.data_types import is_boolean, is_categorical, is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "parallel_coordinates(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(hsl_wheel_colorscale)
                + "\n\n"
                + get_code(make_discrete_colorscale)
                + "\n\n"
                + get_code(get_default_discrete_colorscale)
                + "\n\n"
                + get_code(parallel_coordinates)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the Parallel coordinates section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        parallel_coordinates(df=df, columns=self.columns, color_col=self.color_col)


def parallel_coordinates(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    hide_columns: Optional[List[str]] = None,
    drop_na: bool = False,
    color_col: Optional[str] = None,
    show_colorscale: bool = True,
) -> None:
    """Generate the parallel coordinates interactive plot.

    Parameters
    ----------
    df : pd.DataFrame
        Data for which to generate the parallel coordinates plot.
    columns : List[str], optional
        Columns for which to generate the plot. All columns are used by default.
    hide_columns : List[str], optional
        Columns to exclude from plotting.
    drop_na : bool (default = False)
        Whether to drop NaNs in data.
    color_col : str, optional
        Which column to use for coloring of lines. Can be both numeric and categorical.
    show_colorscale : bool (default = True)
        Whether to show a color scale on the right side of the plot.
    """
    numeric_columns = [col for col in df.columns if is_numeric(df[col])]
    non_numeric_columns = [col for col in df.columns if not is_numeric(df[col])]
    if columns is None:
        non_numeric_columns = [
            col
            for col in non_numeric_columns
            if (is_categorical(df[col]) and df[col].nunique() < 20) or is_boolean(df[col])
        ]
        columns = numeric_columns + non_numeric_columns
    if hide_columns is not None:
        columns = [col for col in columns if col not in hide_columns]
    if drop_na:
        df = df.dropna()
    if color_col is not None:
        is_categorical_color = infer_data_type(df[color_col]) in (
            DataType.CATEGORICAL,
            DataType.UNIQUE,
            DataType.BOOLEAN,
        )

        if is_categorical_color:
            categories = df[color_col].unique()
            colorscale = get_default_discrete_colorscale(n_colors=len(categories))
            # encode categories into numbers
            color_series = pd.Series(pd.Categorical(df[color_col]).codes)
        else:
            color_series = df[color_col]
            colorscale = "Bluered_r"

        line = {
            "color": color_series,
            "colorscale": colorscale,
            "showscale": show_colorscale,
            "colorbar": {"title": color_col, "lenmode": "pixels", "len": 300},
        }

        if is_categorical_color:
            line["colorbar"].update(
                {
                    "tickvals": color_series.unique(),
                    "ticktext": categories,
                    "lenmode": "pixels",
                    "len": min(40 * len(categories), 300),
                }
            )
            # Align the colorbar with the categories
            line.update(
                {
                    "cmin": -0.5,
                    "cmax": len(categories) - 0.5,
                }
            )
    else:
        line = None
    # Add numeric columns to dimensions
    dimensions = [{"label": col_name, "values": df[col_name]} for col_name in numeric_columns]
    # Add categorical columns to dimensions
    for col_name in non_numeric_columns:
        categories = df[col_name].unique()
        values = pd.Series(pd.Categorical(df[col_name]).codes)
        dimensions.append(
            {
                "label": col_name,
                "values": values,
                "tickvals": values.unique(),
                "ticktext": categories,
            }
        )

    fig = go.Figure(go.Parcoords(line=line, dimensions=dimensions))
    fig.show()


class ParallelCategories(Section):
    """
    Generates the Parallel categories subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns for which to generate parallel coordinates.
        All categorical columns with at most `nunique_max` unique values are used by default.
    color_col : str, optional
        Name of column determining colors within categories.
        Both numeric and categorical columns are supported.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
    ):
        self.color_col = color_col
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Parallel categories"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.multivariate_analysis import parallel_categories"]
        return [
            "from edvart.utils import ("
            "    get_default_discrete_colorscale, make_discrete_colorscale, hsl_wheel_colorscale"
            ")",
            "import plotly.graph_objects as go",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "parallel_categories(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(hsl_wheel_colorscale)
                + "\n\n"
                + get_code(make_discrete_colorscale)
                + "\n\n"
                + get_code(get_default_discrete_colorscale)
                + "\n\n"
                + get_code(parallel_categories)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the Parallel coordinates section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        parallel_categories(df=df, columns=self.columns, color_col=self.color_col)


def parallel_categories(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    hide_columns: Optional[List[str]] = None,
    drop_na: bool = False,
    color_col: Optional[str] = None,
) -> None:
    """Generate the parallel coordinates interactive plot.

    Parameters
    ----------
    df : pd.DataFrame
        Data for which to generate the parallel coordinates plot.
    columns : List[str], optional
        Columns for which to generate the plot.
    hide_columns : List[str], optional
        Columns to exclude from plotting.
    drop_na : bool (default = False)
        Whether to drop NaNs in data.
    color_col : str, optional
        Which column to use for coloring of lines. Can be both numeric and categorical.
    """
    if columns is None:
        columns = [
            col
            for col in df.columns
            if is_categorical(df[col]) and df[col].nunique() < 20 or is_boolean(df[col])
        ]
    if hide_columns is not None:
        columns = [col for col in columns if col not in hide_columns]
    if drop_na:
        df = df.dropna()
    if color_col is not None:
        categorical_color = infer_data_type(df[color_col]) in (
            DataType.CATEGORICAL,
            DataType.UNIQUE,
            DataType.BOOLEAN,
        )
        if categorical_color:
            categories = df[color_col].unique()
            colorscale = get_default_discrete_colorscale(n_colors=len(categories))
            # encode categories into numbers
            color_series = pd.Series(pd.Categorical(df[color_col]).codes)
        else:
            color_series = df[color_col]
            colorscale = "Bluered_r"

        line = {
            "color": color_series,
            "colorscale": colorscale,
            "colorbar": {"title": color_col},
        }

        if categorical_color:
            line["colorbar"].update(
                {
                    "tickvals": color_series.unique(),
                    "ticktext": categories,
                    "lenmode": "pixels",
                    "len": min(40 * len(categories), 300),
                }
            )
            # Align the colorbar with the categories
            line.update(
                {
                    "cmin": -0.5,
                    "cmax": len(categories) - 0.5,
                }
            )
    else:
        line = None

    dimensions = [go.parcats.Dimension(values=df[col_name], label=col_name) for col_name in columns]

    fig = go.Figure(go.Parcats(dimensions=dimensions, line=line))
    fig.show()
